/*
 * Name:    mxbl_callback.c
 *
 * Purpose: Handle DCSS and DHS callbacks.
 *
 * Author:  William Lavender
 *
 *--------------------------------------------------------------------------
 *
 * Copyright 2008 Illinois Institute of Technology
 *
 * See the file "LICENSE" for information on usage and redistribution
 * of this file, and for a DISCLAIMER OF ALL WARRANTIES.
 *
 */

#include <stdio.h>
#include <stdlib.h>

#include "mx_util.h"
#include "mx_record.h"
#include "mx_driver.h"
#include "mx_socket.h"
#include "mx_callback.h"
#include "mx_process.h"

#include "mx_analog_input.h"
#include "mx_motor.h"
#include "mx_relay.h"
#include "mx_scaler.h"
#include "mx_timer.h"
#include "mx_variable.h"

#include "mxbl.h"

/*------------------------------------------------------------------*/

mx_status_type
mxbl_setup_record_info( MXBL_EVENT_MANAGER *event_manager,
			MX_RECORD *record,
			char *bluice_name )
{
	static const char fname[] = "mxbl_setup_record_info()";

	MXBL_RECORD_INFO *rinfo;
	mx_status_type mx_status;

	if ( event_manager == (MXBL_EVENT_MANAGER *) NULL ) {
		return mx_error( MXE_NULL_ARGUMENT, fname,
		"The MXBL_EVENT_MANAGER pointer passed was NULL." );
	}
	if ( record == (MX_RECORD *) NULL ) {
		return mx_error( MXE_NULL_ARGUMENT, fname,
		"The MX_RECORD pointer passed was NULL." );
	}

	MX_DEBUG(-2,("%s invoked for record '%s'", fname, record->name));

	if ( record->application_ptr != NULL ) {
		MX_DEBUG(-2,
		("%s: record '%s' already has an MXBL_RECORD_INFO structure.",
			fname, record->name));

		return MX_SUCCESSFUL_RESULT;
	}

	/* Allocate a record info structure for the record. */

	rinfo = (MXBL_RECORD_INFO *) calloc( 1, sizeof(MXBL_RECORD_INFO));

	if ( rinfo == (MXBL_RECORD_INFO *) NULL ) {
		return mx_error( MXE_OUT_OF_MEMORY, fname,
		"Ran out of memory trying to allocate an MXBL_RECORD_INFO "
		"structure for MX record '%s'.", record->name );
	}

	rinfo->event_manager = event_manager;
	rinfo->record = record;

	strlcpy( rinfo->bluice_name,
			bluice_name, MXU_BLUICE_NAME_LENGTH );

	record->application_ptr = rinfo;

	rinfo->rinfo_private = NULL;

	/* Some record types need additional information structures. */

	mx_status = MX_SUCCESSFUL_RESULT;

	switch( record->mx_superclass ) {
	case MXR_DEVICE:
		switch( record->mx_class ) {
		case MXC_TIMER:
			rinfo->rinfo_private =
				calloc( 1, sizeof(MXBL_TIMER_INFO) );

			if ( rinfo->rinfo_private == NULL ) {
				return mx_error( MXE_OUT_OF_MEMORY, fname,
				"Ran out of memory trying to allocate an "
				"MXBL_TIMER_INFO structure for MX record '%s'.",
					record->name );
			}

			mx_status = mxbl_find_ion_chambers( record );
			break;
		}
		break;
	case MXR_SCAN:
		switch( record->mx_type ) {
		case MXS_AD_WEDGE:
			rinfo->rinfo_private =
				calloc( 1, sizeof(MXBL_SCAN_INFO) );

			if ( rinfo->rinfo_private == NULL ) {
				return mx_error( MXE_OUT_OF_MEMORY, fname,
				"Ran out of memory trying to allocate an "
				"MXBL_SCAN_INFO structure for MX record '%s'.",
					record->name );
			}

			mx_status = mxbl_setup_scan_info( record );
			break;
		}
	}

	return mx_status;
}

/*------------------------------------------------------------------*/

mx_status_type
mxbl_register_field_callback( MX_RECORD *record, char *field_name )
{
	static const char fname[] = "mxbl_register_field_callback()";

	MX_RECORD_FIELD *field;
	mx_status_type mx_status;

	if ( record == (MX_RECORD *) NULL ) {
		return mx_error( MXE_NULL_ARGUMENT, fname,
		"The MX_RECORD pointer passed was NULL." );
	}

	MX_DEBUG(-2,("%s invoked for record '%s', field '%s'",
		fname, record->name, field_name ));

	field = mx_get_record_field( record, field_name );

	if ( field == (MX_RECORD_FIELD *) NULL ) {
		return mx_error( MXE_NOT_FOUND, fname,
		"The '%s' field was not found for record '%s'.",
			field_name, record->name );
	}

	mx_status = mx_local_field_add_new_callback( field,
						MXCBT_VALUE_CHANGED,
						mxbl_record_field_callback,
						NULL, NULL );
	return mx_status;
}

/*------------------------------------------------------------------*/

#define MXBL_MESSAGE_TYPE_LENGTH	5

static mx_status_type
mxbl_send_event_update( MX_LIST_ENTRY *list_entry, void *args, void **not_used )
{
	static const char fname[] = "mxbl_send_event_update()";

	MXBL_EVENT *event;
	char *full_message;
	static const char htos[MXBL_MESSAGE_TYPE_LENGTH+1] = "htos_";
	static const char stog[MXBL_MESSAGE_TYPE_LENGTH+1] = "stog_";
	mx_status_type mx_status;

	event = list_entry->list_entry_data;

	if ( event == (MXBL_EVENT *) NULL ) {
		return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
		"The MXBL_EVENT pointer for list entry %p is NULL.",
			list_entry );
	}

	if ( args == NULL ) {
		return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
		"The args pointer for list entry %p is NULL.",
			args );
	}

	full_message = args;

	MX_DEBUG(-2,("%s invoked for socket %d, full_message + 5 = '%s'",
		fname, event->mx_socket->socket_fd, full_message+5));

	switch( event->socket_type ) {
	case MXBLT_DHS_SOCKET:
		memcpy( full_message, htos, MXBL_MESSAGE_TYPE_LENGTH );
		break;
	case MXBLT_DCSS_CLIENT_SOCKET:
		memcpy( full_message, stog, MXBL_MESSAGE_TYPE_LENGTH );
		break;
	default:
		/* Other types of sockets are skipped. */

		return MX_SUCCESSFUL_RESULT;
		break;
	}

	mx_status = mxbl_send_message( event->mx_socket,
				full_message, NULL, 0, MX_BLUICE_PROTOCOL_2 );

	return mx_status;
}

/*------------------------------------------------------------------*/

mx_status_type
mxbl_record_field_callback( MX_CALLBACK *callback, void *args )
{
	static const char fname[] = "mxbl_record_field_callback()";

	MX_RECORD *record;
	MX_RECORD_FIELD *record_field;
	MXBL_RECORD_INFO *rinfo;
	MXBL_TIMER_INFO *timer_info;
	MXBL_EVENT_MANAGER *event_manager;
	mx_bool_type send_value_changed_callback;

	char full_message[200];
	char *message;
	size_t message_sizeof;

	char status[20];

	double motor_position;
	unsigned long motor_status, limit_mask;
	long relay_status;
	MX_TIMER *timer;
	mx_bool_type timer_is_busy;
	long i, num_ion_chambers;
	MX_RECORD **ion_chamber_array;
	MX_RECORD *ion_chamber;
	char value_string[40];
	double ainput_value;
	long scaler_value;
	mx_status_type mx_status;

	if ( callback == (MX_CALLBACK *) NULL ) {
		return mx_error( MXE_NULL_ARGUMENT, fname,
		"The MX_CALLBACK pointer passed was NULL." );
	}

	if ( callback->callback_type != MXCBT_VALUE_CHANGED ) {
		return mx_error( MXE_ILLEGAL_ARGUMENT, fname,
		"Illegal callback type %lu used.  Currently, "
		"MXCBT_VALUE_CHANGED (%d) callbacks are allowed here.",
			callback->callback_type, MXCBT_VALUE_CHANGED );
	}

	record_field = callback->u.record_field;

	if ( record_field == (MX_RECORD_FIELD *) NULL ) {
		return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
		"No MX_RECORD_FIELD pointer was specified for callback %p",
			callback );
	}

	record = record_field->record;

	if ( record == (MX_RECORD *) NULL ) {
		return mx_error( MXE_UNSUPPORTED, fname,
		"Callbacks are not supported for temporary record fields.  "
		"Field name = '%s'", record_field->name );
	}

	rinfo = record->application_ptr;

	if ( rinfo == (MXBL_RECORD_INFO *) NULL ) {
		return mx_error( MXE_INITIALIZATION_ERROR, fname,
		"No MXBL_RECORD_INFO structure has been allocated "
		"for record '%s'.", record->name );
	}

	if ( callback->get_new_value ) {
		/* Process the record field to get the new value. */

#if 0
		MX_DEBUG(-2,("%s: Processing '%s.%s'",
			fname, record->name, record_field->name));
#endif

		mx_status = mx_process_record_field( record,
						record_field, MX_PROCESS_GET,
						&send_value_changed_callback );

		if ( mx_status.code != MXE_SUCCESS )
			return mx_status;
	} else {
		/* Send the current value of the field. */

		send_value_changed_callback = TRUE;
	}

#if 0
	MX_DEBUG(-2,("%s: '%s.%s' send_value_changed_callback = %d",
		fname, record->name, record_field->name,
		send_value_changed_callback));
#endif

	if ( send_value_changed_callback == FALSE ) {
		return MX_SUCCESSFUL_RESULT;
	}

	MX_DEBUG(-2,("%s: sending value changed callback for '%s.%s'",
		fname, record->name, record_field->name));

	/* If we get here, then a value has changed.  We must now report
	 * the new value(s) back to the calling process.
	 */

	/* Skip over htos_ or stog_. */

	message = full_message + MXBL_MESSAGE_TYPE_LENGTH;

	message_sizeof = sizeof(full_message) - MXBL_MESSAGE_TYPE_LENGTH;

	message[0] = '\0';

	switch( record->mx_superclass ) {
	case MXR_DEVICE:
	    switch( record->mx_class ) {
	    case MXC_MOTOR:
		switch( record_field->label_value ) {
		case MXLV_MTR_GET_EXTENDED_STATUS:
		    mx_status = mx_motor_get_extended_status( record,
					&motor_position, &motor_status );

		    if ( mx_status.code != MXE_SUCCESS )
			return mx_status;

		    limit_mask = MXSF_MTR_POSITIVE_LIMIT_HIT
		    			| MXSF_MTR_NEGATIVE_LIMIT_HIT;

		    /*---*/

		    /* Note: The 'aborted' status is only sent by the
		     * stoh_abort_all handler.
		     */

		    if ( (motor_status & limit_mask) == limit_mask ) {

		    	strlcpy( status, "both_hw_limits", sizeof(status) );
		    } else
		    if ( motor_status & MXSF_MTR_POSITIVE_LIMIT_HIT ) {

		    	strlcpy( status, "cw_hw_limit", sizeof(status) );
		    } else
		    if ( motor_status & MXSF_MTR_NEGATIVE_LIMIT_HIT ) {

		    	strlcpy( status, "ccw_hw_limit", sizeof(status) );
		    } else
		    if ( motor_status & MXSF_MTR_ERROR_BITMASK ) {

		    	strlcpy( status, "unknown", sizeof(status) );
		    } else
		    if ( motor_status & MXSF_MTR_IS_BUSY ) {

		    	strlcpy( status, "moving", sizeof(status) );
		    } else {
		    	strlcpy( status, "normal", sizeof(status) );
		    }

		    /*---*/

		    if ( rinfo->operation_in_progress ) {
			if ( motor_status & MXSF_MTR_IS_BUSY ) {
			    snprintf( message, message_sizeof,
				"update_motor_position %s %f %s",
					record->name, motor_position, status );
			} else {
			    snprintf( message, message_sizeof,
				"motor_move_completed %s %f %s",
					record->name, motor_position, status );

			    rinfo->operation_in_progress = FALSE;
			}
		    } else {
			if ( motor_status & MXSF_MTR_IS_BUSY ) {
			    snprintf( message, message_sizeof,
				"motor_move_started %s %f",
					record->name, motor_position );

			    rinfo->operation_in_progress = TRUE;
			} else {
			    /* If we get here, then no motor move was
			     * previously in progress and no new motor
			     * move has been started.  In this case,
			     * we do not send a message to DCSS.
			     */

			    return MX_SUCCESSFUL_RESULT;
			}
		    }
		    break;
		}
		break;

	    case MXC_RELAY:
		switch( record_field->label_value ) {
		case MXLV_RLY_RELAY_STATUS:
		    mx_status = mx_get_relay_status( record, &relay_status );

		    if ( mx_status.code != MXE_SUCCESS )
		    	return mx_status;

		    switch( relay_status ) {
		    case MXF_RELAY_IS_CLOSED:
		    	snprintf( message, message_sizeof,
				"report_shutter_state %s closed",
					record->name );
			break;
		    case MXF_RELAY_IS_OPEN:
		    	snprintf( message, message_sizeof,
				"report_shutter_state %s open",
					record->name );
		    	break;
		    case MXF_RELAY_ILLEGAL_STATUS:
		    default:
			/* The safer thing to do here is to report 
			 * the shutter status as open, so that people
			 * will not accidentally think that X-rays
			 * are turned off.
			 */
		    	snprintf( message, message_sizeof,
				"report_shutter_state %s open",
					record->name );
			mx_warning( "MX relay record '%s' reports itself "
			"as being in illegal state %ld",
				record->name, relay_status );
		    	break;
		    }
		    break;
		}
		break;

	    case MXC_TIMER:
		switch( record_field->label_value ) {
		case MXLV_TIM_BUSY:
		    mx_status = mx_timer_is_busy( record, &timer_is_busy );

		    if ( mx_status.code != MXE_SUCCESS )
		    	return mx_status;

		    if ( rinfo->operation_in_progress == FALSE ) {
		    	if ( timer_is_busy ) {
			    /* The timer was not previously busy, but now
			     * _is_ busy, so record that fact.
			     */

			    rinfo->operation_in_progress = TRUE;

			    return MX_SUCCESSFUL_RESULT;
			} else {
			    /* The timer was not previously busy and is not
			     * busy now, so there is no need to send a
			     * message to the caller at this time.
			     */

			    return MX_SUCCESSFUL_RESULT;
		    	}
		    } else {
		    	if ( timer_is_busy ) {
			    /* The timer is still counting down the
			     * measurement time, so there is no need
			     * to send a message to the caller at
			     * this time.
			     */

			    return MX_SUCCESSFUL_RESULT;
			} else {
			    /* The timer _was_ previously busy and now
			     * is _not_ busy, so it is time to send a
			     * message containing the ion chamber readings
			     * to the caller.
			     */

			    timer = record->record_class_struct;

			    snprintf(message, message_sizeof,
			    	"report_ion_chambers %f",
				timer->last_measurement_time );

			    timer_info = rinfo->rinfo_private;

			    num_ion_chambers = timer_info->num_ion_chambers;
			    ion_chamber_array = timer_info->ion_chamber_array;

			    for ( i = 0; i < num_ion_chambers; i++ ) {
			    	ion_chamber = ion_chamber_array[i];

				snprintf(value_string, sizeof(value_string),
					" %s", ion_chamber->name);

				strlcat(message, value_string, message_sizeof);

				switch( ion_chamber->mx_class ) {
				case MXC_ANALOG_INPUT:
				    mx_status = mx_analog_input_read(
						  ion_chamber, &ainput_value );

				    if ( mx_status.code != MXE_SUCCESS )
				    	return mx_status;

				    snprintf(value_string, sizeof(value_string),
				    	" %f", ainput_value);
				    break;

				case MXC_SCALER:
				    mx_status = mx_scaler_read( ion_chamber,
				    				&scaler_value );

				    if ( mx_status.code != MXE_SUCCESS )
				    	return mx_status;

				    snprintf(value_string, sizeof(value_string),
				    	" %ld", scaler_value);
				    break;
				}
				strlcat(message, value_string, message_sizeof);
			    }
			    MX_DEBUG(-2,("%s: ion chamber message = '%s'",
			    	fname, message));
		    	}
		    }
		}
	    	break;
	    }
	    break;
	}

	if ( strlen(message) == 0 ) {
		return mx_error( MXE_UNSUPPORTED, fname,
		"There is no defined Blu-Ice response mechanism for "
		"the MX record field '%s.%s'",
			record->name, record_field->name );
	}

	/* Walk through all of the active DCSS and DHS event handlers
	 * and send the updates back to them.
	 */

	event_manager = rinfo->event_manager;

	if ( event_manager == (MXBL_EVENT_MANAGER *) NULL ) {
		return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
		"The MXBL_EVENT_MANAGER pointer for record '%s' is NULL.",
			record->name );
	}

	MX_DEBUG(-2,("%s: message = '%s'", fname, message));

	mx_status = mx_list_traverse( event_manager->event_list,
				mxbl_send_event_update, full_message, NULL );

	return mx_status;
}

