/*
 * Name:    mxbl_handlers.c
 *
 * Purpose: Event handlers used by both DCSS and DHS.
 *
 * Author:  William Lavender
 *
 *--------------------------------------------------------------------------
 *
 * Copyright 2008 Illinois Institute of Technology
 *
 * See the file "LICENSE" for information on usage and redistribution
 * of this file, and for a DISCLAIMER OF ALL WARRANTIES.
 *
 */

#include <stdio.h>
#include <stdlib.h>
#include <ctype.h>

#include "mx_util.h"
#include "mx_record.h"
#include "mx_driver.h"
#include "mx_list.h"
#include "mx_socket.h"
#include "mx_callback.h"
#include "mx_process.h"

#include "mx_scaler.h"
#include "mx_scan.h"
#include "mx_bluice.h"
#include "v_bluice_string.h"

#include "mx_poison.h"

#include "mxbl.h"

/*------------------------------------------------------------------*/

mx_status_type
mxblh_abort_all( MXBL_EVENT *event )
{
	static const char fname[] = "mxblh_abort_all()";

	MX_SOCKET *mx_socket;
	MX_RECORD *record_list, *current_record;
	char *ptr, *token_ptr, *abort_type;
	mx_bool_type immediate_abort;

	MX_DEBUG(-2,("%s invoked.", fname));

	mx_socket = event->mx_socket;

	record_list = event->event_manager->record_list;

	/* Skip over the command name. */

	ptr = event->message_buffer;

	token_ptr = mx_string_token( &ptr, " " );

	if ( token_ptr == NULL ) {
		return mx_error( MXE_UNPARSEABLE_STRING, fname,
		"The DHS message buffer received for socket %d was empty.  "
		"This should not be able to happen.", mx_socket->socket_fd );
	}

	/* The next token should be the type of abort to perform. */

	abort_type = mx_string_token( &ptr, " " );

	if ( abort_type == NULL ) {
		return mx_error( MXE_UNPARSEABLE_STRING, fname,
		"No abort type was seen in the stoh_abort_all "
		"command received by DHS socket %d.",
			mx_socket->socket_fd );
	}

	/* Now do the abort. */

	if ( strcmp( abort_type, "soft" ) == 0 ) {
		MX_DEBUG(-2,("%s: Soft abort requested.", fname));

		immediate_abort = FALSE;
	} else
	if ( strcmp( abort_type, "hard" ) == 0 ) {
		MX_DEBUG(-2,("%s: Hard abort requested.", fname));

		immediate_abort = TRUE;
	} else {
		return mx_error( MXE_NETWORK_IO_ERROR, fname,
		"Unrecognized abort type '%s' requested by the stoh_abort_all "
		"command sent by Blu-Ice socket %d.",
			abort_type, mx_socket->socket_fd );
	}

	current_record = record_list->next_record;

	while( current_record != record_list ) {
		if ( current_record->mx_class == MXC_MOTOR ) {

			MX_DEBUG(-2,("%s: Sending abort to '%s'",
				fname, current_record->name));

			if ( immediate_abort ) {
				mx_motor_immediate_abort( current_record );
			} else {
				mx_motor_soft_abort( current_record );
			}
		}
		current_record = current_record->next_record;
	}

	MX_DEBUG(-2,("%s complete.", fname));

	return MX_SUCCESSFUL_RESULT;
}

/*------------------------------------------------------------------*/

#define MXBL_IC_OFFSET	3      /* The list of ion chambers starts at argv[3]. */

mx_status_type
mxblh_read_ion_chambers( MXBL_EVENT *event )
{
	static const char fname[] = "mxblh_read_ion_chambers()";

	MX_SOCKET *mx_socket;
	MX_RECORD *record_list, *ion_chamber_record, *timer_record;
	MX_ANALOG_INPUT *ainput;
	MX_SCALER *scaler;
	MXBL_RECORD_INFO *record_info;
	int argc;
	char **argv;
	char *endptr;
	unsigned long i, repeat_count;
	double measurement_time;
	mx_status_type mx_status;

	MX_DEBUG(-2,("%s invoked.", fname));

	mx_socket = event->mx_socket;

	record_list = event->event_manager->record_list;

	mx_string_split( event->message_buffer, " ", &argc, &argv );

	if ( argc <= MXBL_IC_OFFSET ) {
		mx_status = mx_error( MXE_NETWORK_IO_ERROR, fname,
		"No ion chambers were specified in the '%s' command "
		"received by socket %d.", argv[0], mx_socket->socket_fd );

		mx_free(argv);
		return mx_status;
	}

	/* Get the measurement time. */

	measurement_time = strtod( argv[1], &endptr );

	if ( *endptr != '\0' ) {
		mx_status = mx_error( MXE_ILLEGAL_ARGUMENT, fname,
			"The measurement time '%s' requested by socket %d "
			"for ion chamber '%s' is not a legal floating point "
			"number.", argv[1], mx_socket->socket_fd, argv[3] );

		mx_free(argv);
		return mx_status;
	}

	/* Get the repeat count. */

	repeat_count = mx_string_to_unsigned_long( argv[2] );

	if ( repeat_count != 0 ) {
		mx_warning( "The MX Blu-Ice server does not currently "
		"support repeat counts for the stoh_read_ion_chambers "
		"command." );
	}

	/* Walk through the list of ion chambers. */

	MX_DEBUG(-2,("%s: %d ion chambers requested.",
		fname, argc - MXBL_IC_OFFSET));

	timer_record = NULL;

	for ( i = MXBL_IC_OFFSET; i < argc; i++ ) {

		/* Find the ion chamber in the MX database. */

		ion_chamber_record = mx_get_record( record_list, argv[i] );

		if ( ion_chamber_record == (MX_RECORD *) NULL ) {
			mx_status = mx_error( MXE_NOT_FOUND, fname,
				"Ion chamber '%s' requested by socket %d "
				"is not found in the MX database.",
					argv[i], mx_socket->socket_fd );

			mx_free(argv);
			return mx_status;
		}

		switch( ion_chamber_record->mx_class ) {
		case MXC_ANALOG_INPUT:
			ainput = ion_chamber_record->record_class_struct;

			if ( ainput->timer_record == NULL ) {
				mx_free(argv);
				return mx_error( MXE_ILLEGAL_ARGUMENT, fname,
				"Ion chamber '%s' does not have an "
				"associated timer.", ion_chamber_record->name );
			} else
			if ( timer_record == NULL ) {
				timer_record = ainput->timer_record;
			} else
			if ( timer_record != ainput->timer_record ) {
				mx_status = mx_error(MXE_ILLEGAL_ARGUMENT,fname,
				"Ion chamber '%s' uses a timer '%s' that "
				"is not the same as the timer '%s' used "
				"by the first ion chamber '%s'.",
					ion_chamber_record->name,
					ainput->timer_record->name,
					timer_record->name,
					argv[MXBL_IC_OFFSET] );

				mx_free(argv);
				return mx_status;
			}
			break;

		case MXC_SCALER:
			scaler = ion_chamber_record->record_class_struct;

			if ( scaler->timer_record == NULL ) {
				mx_free(argv);
				return mx_error( MXE_ILLEGAL_ARGUMENT, fname,
				"Ion chamber '%s' does not have an "
				"associated timer.", ion_chamber_record->name );
			} else
			if ( timer_record == NULL ) {
				timer_record = scaler->timer_record;

			} else
			if ( timer_record != scaler->timer_record ) {
				mx_status = mx_error(MXE_ILLEGAL_ARGUMENT,fname,
				"Ion chamber '%s' uses a timer '%s' that "
				"is not the same as the timer '%s' used "
				"by the first ion chamber '%s'.",
					ion_chamber_record->name,
					scaler->timer_record->name,
					timer_record->name,
					argv[MXBL_IC_OFFSET] );

				mx_free(argv);
				return mx_status;
			}

			/* Clear the scaler. */

			mx_status = mx_scaler_clear( ion_chamber_record );

			if ( mx_status.code != MXE_SUCCESS ) {
				mx_free(argv);

				return mx_error( mx_status.code, fname,
				"Unable to clear scaler '%s'.",
					ion_chamber_record->name );
			}

			break;

		default:
			mx_free(argv);

			return mx_error( MXE_ILLEGAL_ARGUMENT, fname,
			"MX record '%s' of type '%s' cannot be used as "
			"a Blu-Ice ion chamber.",
				ion_chamber_record->name,
				mx_get_driver_name( ion_chamber_record ) );
			break;
		}
	}

	mx_free(argv);

	if ( timer_record == (MX_RECORD *) NULL ) {
		return mx_error( MXE_SOFTWARE_CONFIGURATION_ERROR, fname,
		"None of the ion chambers requested by socket %d "
		"have associated timers.", mx_socket->socket_fd );
	}

	record_info = timer_record->application_ptr;

	if ( record_info == (MXBL_RECORD_INFO *) NULL ) {
		return mx_error( MXE_INITIALIZATION_ERROR, fname,
		"The timer record '%s' does not have an associated "
		"MXBL_RECORD_INFO structure.", timer_record->name );
	}

	/* Start the timer. */

	mx_status = mx_timer_start( timer_record, measurement_time );

	if ( mx_status.code != MXE_SUCCESS )
		return mx_status;

	record_info->operation_in_progress = TRUE;

	MX_DEBUG(-2,("%s complete.", fname));

	return MX_SUCCESSFUL_RESULT;
}

/*------------------------------------------------------------------*/

mx_status_type
mxblh_set_shutter_state( MXBL_EVENT *event )
{
	static const char fname[] = "mxblh_set_shutter_state()";

	MX_SOCKET *mx_socket;
	MX_RECORD *record_list, *shutter_record;
	MXBL_RECORD_INFO *rinfo;
	int argc;
	char **argv;
	mx_status_type mx_status;

	MX_DEBUG(-2,("%s invoked.", fname));

	mx_socket = event->mx_socket;

	record_list = event->event_manager->record_list;

	mx_string_split( event->message_buffer, " ", &argc, &argv );

	if ( argc != 3 ) {
		mx_status = mx_error( MXE_NETWORK_IO_ERROR, fname,
			"A '%s' message was received by socket %d that has "
			"%d arguments when the correct number of arguments "
			"is 3.", argv[0], mx_socket->socket_fd, argc );

		mx_free(argv);
		return mx_status;
	}

	/* Find the shutter in the MX database. */

	shutter_record = mx_get_record( record_list, argv[1] );

	if ( shutter_record == (MX_RECORD *) NULL ) {
		mx_status = mx_error( MXE_NOT_FOUND, fname,
			"DHS socket %d attempted to change the state of "
			"shutter '%s', but there is no shutter by that name.",
				mx_socket->socket_fd, argv[1] );
		mx_free(argv);
		return mx_status;
	}

	rinfo = shutter_record->application_ptr;

	if ( rinfo == (MXBL_RECORD_INFO *) NULL ) {
		mx_status = mx_error( MXE_INITIALIZATION_ERROR, fname,
			"No MXBL_RECORD_INFO structure has been allocated "
			"for record '%s'.", shutter_record->name );
		mx_free(argv);
		return mx_status;
	}

	/* Change the shutter state. */

	MX_DEBUG(-2,("%s: Changing the state of '%s' to '%s'",
		fname, shutter_record->name, argv[2]));

	if ( strcmp( argv[2], "open" ) == 0 ) {
		mx_status = mx_relay_command( shutter_record, MXF_OPEN_RELAY );
	} else
	if ( strcmp( argv[2], "closed" ) == 0 ) {
		mx_status = mx_relay_command( shutter_record, MXF_CLOSE_RELAY );
	} else {
		mx_status = mx_error( MXE_NETWORK_IO_ERROR, fname,
		"An unrecognized shutter state '%s' was requested by "
		"DHS socket %d for shutter '%s'.",
			argv[2], mx_socket->socket_fd,
			shutter_record->name );
	}

	mx_free(argv);

	return mx_status;
}

/*------------------------------------------------------------------*/

mx_status_type
mxblh_start_motor_move( MXBL_EVENT *event )
{
	static const char fname[] = "mxblh_start_motor_move()";

	MX_SOCKET *mx_socket;
	MX_RECORD *record_list, *motor_record;
	MXBL_RECORD_INFO *rinfo;
	int argc;
	char **argv;
	char *endptr;
	double destination;
	char message[80];
	mx_status_type mx_status;

	MX_DEBUG(-2,("%s invoked.", fname));

	mx_socket = event->mx_socket;

	record_list = event->event_manager->record_list;

	mx_string_split( event->message_buffer, " ", &argc, &argv );

	if ( argc != 3 ) {
		mx_status = mx_error( MXE_NETWORK_IO_ERROR, fname,
			"A '%s' message was received by socket %d that has "
			"%d arguments when the correct number of arguments "
			"is 3.", argv[0], mx_socket->socket_fd, argc );

		mx_free(argv);
		return mx_status;
	}

	/* Find the motor in the MX database. */

	motor_record = mx_get_record( record_list, argv[1] );

	if ( motor_record == (MX_RECORD *) NULL ) {
		mx_status = mx_error( MXE_NOT_FOUND, fname,
		"DHS socket %d requested a move for motor '%s', but "
		"there is no motor by that name.",
			mx_socket->socket_fd, argv[1] );

		mx_free(argv);
		return mx_status;
	}

	destination = strtod( argv[2], &endptr );

	if ( *endptr != '\0' ) {
		mx_status = mx_error( MXE_ILLEGAL_ARGUMENT, fname,
		"The destination '%s' requested by DHS socket %d "
		"for motor '%s' is not a legal floating point number.",
			argv[2], mx_socket->socket_fd, motor_record->name );

		mx_free(argv);
		return mx_status;
	}

	mx_free(argv);		/* We do not need argv anymore. */

	rinfo = motor_record->application_ptr;

	if ( rinfo == (MXBL_RECORD_INFO *) NULL ) {
		return mx_error( MXE_INITIALIZATION_ERROR, fname,
		"No MXBL_RECORD_INFO structure has been allocated "
		"for record '%s'.", motor_record->name );
	}

	/* Start the move. */

	MX_DEBUG(-2,("%s: Moving '%s' to %g",
		fname, motor_record->name, destination));

	mx_status = mx_motor_move_absolute( motor_record, destination, 0 );

	if ( mx_status.code != MXE_SUCCESS ) {
		return mx_status;
	}

	rinfo->operation_in_progress = TRUE;

	/* Send a message back to the caller telling it that
	 * the motor move has started.
	 */

	switch( event->socket_type ) {
	case MXBLT_DHS_SOCKET:
		snprintf( message, sizeof(message),
			"htos_motor_move_started %s %f",
			motor_record->name, destination );
		break;
	case MXBLT_DCSS_CLIENT_SOCKET:
		snprintf( message, sizeof(message),
			"stog_motor_move_started %s %f",
			motor_record->name, destination );
		break;
	default:
		snprintf( message, sizeof(message),
			"error_motor_move_started %s %f",
			motor_record->name, destination );
		break;
	}

	mx_status = mxbl_send_message( mx_socket, message,
					NULL, 0, MX_BLUICE_PROTOCOL_2 );
	return mx_status;
}

/*------------------------------------------------------------------*/

static void
mxblp_free_ion_chambers( MX_RECORD **ion_chamber_array,
				MX_RECORD *timer_record )
{
	MXBL_RECORD_INFO *record_info;
	MXBL_TIMER_INFO *timer_info;

	mx_free( ion_chamber_array );

	if ( timer_record != NULL ) {
		record_info = timer_record->application_ptr;

		if ( record_info != NULL ) {
			timer_info = record_info->rinfo_private;

			if ( timer_info != NULL ) {
				timer_info->num_ion_chambers = 0;

				mx_free( timer_info->ion_chamber_array );
			}
		}
	}
}

mx_status_type
mxbl_find_ion_chambers( MX_RECORD *timer_record )
{
	static const char fname[] = "mxbl_find_ion_chambers()";

	MX_RECORD *record_list, *current_record;
	MX_ANALOG_INPUT *ainput;
	MX_SCALER *scaler;
	MXBL_RECORD_INFO *record_info;
	MXBL_TIMER_INFO *timer_info;
	long num_ion_chambers, num_blocks, block_size, array_size;
	MX_RECORD **ion_chamber_array, **new_ion_chamber_array;

	if ( timer_record == (MX_RECORD *) NULL ) {
		return mx_error( MXE_NULL_ARGUMENT, fname,
		"The timer record pointer passed was NULL." );
	}

	MX_DEBUG(-2,("%s invoked for timer '%s'.", fname, timer_record->name));

	record_list = timer_record->list_head;

	/* Allocate an initial array of ion chamber records. */

	num_blocks = 1;
	block_size = 5;

	array_size = num_blocks * block_size;

	ion_chamber_array = (MX_RECORD **)
		malloc( array_size * sizeof(MX_RECORD *) );

	if ( ion_chamber_array == (MX_RECORD **) NULL ) {
		return mx_error( MXE_OUT_OF_MEMORY, fname,
		"Ran out of memory trying to allocate a %ld-element "
		"array of ion chamber records.", array_size );
	}

	num_ion_chambers = 0;

	/* Walk through the MX database looking for analog input
	 * and scaler records that use this timer.
	 */

	current_record = record_list->next_record;

	while( current_record != record_list ) {

		if ( num_ion_chambers >= array_size ) {
			num_blocks++;

			array_size = num_blocks * block_size;

			MX_DEBUG(-2,("%s: realloc()-ing to %ld",
				fname, array_size));

			new_ion_chamber_array = realloc( ion_chamber_array,
								array_size );

			if ( new_ion_chamber_array == NULL ) {
				mxblp_free_ion_chambers( ion_chamber_array,
							timer_record );

				return mx_error( MXE_OUT_OF_MEMORY, fname,
				"Ran out of memory trying to increase the "
				"number of records in the ion chamber array "
				"to %ld.", array_size );
			}

			ion_chamber_array = new_ion_chamber_array;
		}

		/* See if this record is used by the timer record. */

		switch( current_record->mx_class ) {
		case MXC_ANALOG_INPUT:
			ainput = current_record->record_class_struct;

			if ( ainput->timer_record == timer_record ) {
				ion_chamber_array[num_ion_chambers]
					= current_record;

				num_ion_chambers++;
			}
			break;

		case MXC_SCALER:
			scaler = current_record->record_class_struct;

			if ( scaler->timer_record == timer_record ) {
				ion_chamber_array[num_ion_chambers]
					= current_record;

				num_ion_chambers++;
			}
			break;

		default:
			break;
		}

		current_record = current_record->next_record;
	}

#if 1
	{
		long i;

		MX_DEBUG(-2,("%s: num_ion_chambers = %ld",
			fname, num_ion_chambers));

		for ( i = 0; i < num_ion_chambers; i++ ) {
			MX_DEBUG(-2,("%s:   ion_chamber[%ld] = '%s'",
				fname, i, ion_chamber_array[i]->name ));
		}
	}
#endif
	/* Attach the array of ion chambers to the MXBL_TIMER_INFO structure
	 * of the timer record.
	 */

	if ( timer_record->application_ptr != NULL ) {
		record_info = timer_record->application_ptr;
	} else {
		record_info = (MXBL_RECORD_INFO *)
					calloc( 1, sizeof(MXBL_RECORD_INFO) );

		if ( record_info == (MXBL_RECORD_INFO *) NULL ) {
			mxblp_free_ion_chambers( ion_chamber_array,
						timer_record );

			return mx_error( MXE_OUT_OF_MEMORY, fname,
			"Ran out of memory trying to allocate an "
			"MXBL_RECORD_INFO structure for timer '%s'.",
				timer_record->name );
		}

		timer_record->application_ptr = record_info;
	}

	if ( record_info->rinfo_private != NULL ) {
		timer_info = record_info->rinfo_private;
	} else {
		timer_info = (MXBL_TIMER_INFO *)
					calloc( 1, sizeof(MXBL_TIMER_INFO) );

		if ( timer_info == (MXBL_TIMER_INFO *) NULL ) {
			mxblp_free_ion_chambers( ion_chamber_array,
						timer_record );

			return mx_error( MXE_OUT_OF_MEMORY, fname,
			"Ran out of memory trying to allocate an "
			"MXBL_TIMER_INFO structure for timer '%s'.",
				timer_record->name );
		}

		record_info->rinfo_private = timer_info;
	}

	mx_free( timer_info->ion_chamber_array );

	timer_info->num_ion_chambers  = num_ion_chambers;
	timer_info->ion_chamber_array = ion_chamber_array;

	MX_DEBUG(-2,("%s complete.", fname));

	return MX_SUCCESSFUL_RESULT;
}

/*------------------------------------------------------------------*/

mx_status_type
mxbl_setup_scan_info( MX_RECORD *record )
{
	static const char fname[] = "mxbl_setup_scan_info()";

	MX_SCAN *scan;
	MXBL_RECORD_INFO *rinfo;
	MXBL_SCAN_INFO *scan_info;
	char datafile_fullpath[MXU_FILENAME_LENGTH+1];
	char *ptr, *filename_ptr;
	int num_items;

    	scan = record->record_superclass_struct;

	if ( scan == (MX_SCAN *) NULL ) {
		return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
		"The MX_SCAN pointer for record '%s' is NULL.",
			record->name );
	}

	rinfo = record->application_ptr;

	if ( rinfo == (MXBL_RECORD_INFO *) NULL ) {
		return mx_error( MXE_INITIALIZATION_ERROR, fname,
		"The MXBL_RECORD_INFO pointer for record '%s' is NULL.",
			record->name );
	}

	scan_info = rinfo->rinfo_private;

	if ( scan_info == (MXBL_SCAN_INFO *) NULL ) {
		return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
		"The MXBL_SCAN_INFO pointer for record '%s' is NULL.",
			record->name );
	}

	strlcpy( scan_info->run_status, "complete",
			sizeof(scan_info->run_status) );

	/*-----------------------------------------------*/

	/* The run label is the number at the end of the MX scan record name. */

        num_items = sscanf( record->name, "run%lu",
				&(scan_info->run_label) );

	if ( num_items != 1 ) {
		return mx_error( MXE_ILLEGAL_ARGUMENT, fname,
		"Scan '%s' does not fit the Blu-Ice pattern of "
		"'runNNN' where NNN is the run number.",
			record->name );
	}

	/*-----------------------------------------------*/

	/* Split up the scan's datafile_name field into
	 * the parts needed by Blu-Ice.
	 */

	MX_DEBUG(-2,("%s: scan->datafile_name = '%s'",
		fname, scan->datafile_name));

	strlcpy( datafile_fullpath, scan->datafile_name,
			sizeof(datafile_fullpath) );

	/* Look for the last '/' character in the filename. */

	ptr = strrchr( datafile_fullpath, '/' );

	if ( ptr == NULL ) {
		mx_get_current_directory_name( scan_info->directory,
					sizeof(scan_info->directory) );

		filename_ptr = datafile_fullpath;
	} else {
		filename_ptr = ptr+1;

		*ptr = '\0';

		strlcpy( scan_info->directory, datafile_fullpath,
					sizeof(scan_info->directory) );
	}

	strlcpy( scan_info->file_root, filename_ptr,
			sizeof(scan_info->file_root) );

	MX_DEBUG(-2,("%s: directory = '%s', file_root = '%s'",
		fname, scan_info->directory, scan_info->file_root));

	MX_DEBUG(-2,("%s: *** FIXME *** Compute a real value for next_frame",
		fname));

	scan_info->next_frame = 1;
	scan_info->start_frame_label = 1;

	return MX_SUCCESSFUL_RESULT;
}

/*------------------------------------------------------------------*/

mx_status_type
mxbl_set_string( MXBL_EVENT_MANAGER *event_manager, 
		MX_RECORD *string_record,
		char *string_value )
{
	static const char fname[] = "mxbl_set_string()";

	MX_BLUICE_STRING *bluice_string;
	long num_dimensions, field_type;
	long *dimension;
	void *pointer_to_value;
	char *string_ptr;
	char message[500];
	mx_status_type mx_status;

	if ( event_manager == (MXBL_EVENT_MANAGER *) NULL ) {
		return mx_error( MXE_NULL_ARGUMENT, fname,
		"The MXBL_EVENT_MANAGER pointer passed was NULL." );
	}
	if ( string_record == (MX_RECORD *) NULL ) {
		return mx_error( MXE_NULL_ARGUMENT, fname,
		"The MX_RECORD pointer passed was NULL." );
	}
	if ( string_value == (char *) NULL ) {
		return mx_error( MXE_NULL_ARGUMENT, fname,
		"The string pointer passed was NULL." );
	}

	if ( string_record->mx_type != MXV_BLUICE_SELF_STRING ) {
		return mx_error( MXE_ILLEGAL_ARGUMENT, fname,
		"Record '%s' is not a 'bluice_self_string' record.",
			string_record->name );
	}

	mx_status = mx_get_variable_parameters( string_record,
			&num_dimensions, &dimension,
			&field_type, &pointer_to_value );

	if ( mx_status.code != MXE_SUCCESS )
		return mx_status;

	if ( (num_dimensions != 1) || (field_type != MXFT_STRING) ) {
		return mx_error( MXE_ILLEGAL_ARGUMENT, fname,
		"Record '%s' is not a 1-dimensional string record.",
			string_record->name );
	}

	MX_DEBUG(-2,("%s: dimension[0] = %lu", fname, dimension[0]));

	bluice_string = string_record->record_type_struct;

	if ( bluice_string == (MX_BLUICE_STRING *) NULL ) {
		return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
		"The MX_BLUICE_STRING pointer for record '%s' is NULL.",
			string_record->name );
	}

	/* Copy the string value to the string record. */

	string_ptr = pointer_to_value;

	strlcpy( string_ptr, string_value, dimension[0] - 1 );

	/* Send an stog_set_string_completed message to all of the clients. */

	snprintf( message, sizeof(message),
	"stog_set_string_completed %s normal %s",
		bluice_string->bluice_name, string_ptr );

	mx_status = mxbl_broadcast_dcss_message( event_manager, message );

	return mx_status;
}

/*------------------------------------------------------------------*/

static MX_RECORD *mxblp_collect_msg_record = NULL;

mx_status_type
mxbl_set_collect_msg( MXBL_EVENT_MANAGER *event_manager,
			char *string_value )
{
	static const char fname[] = "mxbl_set_collect_msg()";

	mx_status_type mx_status;

	if ( event_manager == (MXBL_EVENT_MANAGER *) NULL ) {
		return mx_error( MXE_NULL_ARGUMENT, fname,
		"The MXBL_EVENT_MANAGER pointer passed was NULL." );
	}
	if ( string_value == (char *) NULL ) {
		return mx_error( MXE_NULL_ARGUMENT, fname,
		"The string pointer passed was NULL." );
	}

	if ( mxblp_collect_msg_record == (MX_RECORD *) NULL ) {
		mxblp_collect_msg_record = mx_get_record(
				event_manager->record_list, "collect_msg" );
	}

	mx_status = mxbl_set_string( event_manager,
				mxblp_collect_msg_record,
				string_value );

	return mx_status;
}


