/*
 * Name:    mxbl_message.c
 *
 * Purpose: Send and receive Blu-Ice messages.
 *
 * Author:  William Lavender
 *
 *--------------------------------------------------------------------------
 *
 * Copyright 2008 Illinois Institute of Technology
 *
 * See the file "LICENSE" for information on usage and redistribution
 * of this file, and for a DISCLAIMER OF ALL WARRANTIES.
 *
 */

#include <stdio.h>

#include "mx_util.h"
#include "mx_record.h"
#include "mx_list.h"
#include "mx_callback.h"
#include "mx_socket.h"
#include "mx_bluice.h"

#include "mx_poison.h"

#include "mxbl.h"

/*------------------------------------------------------------------*/

mx_status_type
mxbl_send_message( MX_SOCKET *mx_socket,
		char *text_data,
		char *binary_data,
		long binary_data_length,
		long protocol_version )
{
	static const char fname[] = "mxbl_send_message()";

	char message_header[MX_BLUICE_MSGHDR_LENGTH+1];
	char null_pad[500];
	mx_status_type mx_status;

	long text_data_length, bytes_sent, null_bytes_to_send;

	if ( mx_socket == (MX_SOCKET *) NULL ) {
		return mx_error( MXE_NULL_ARGUMENT, fname,
		"The MX_SOCKET pointer passed was NULL." );
	}
	if ( text_data == (char *) NULL ) {
		return mx_error( MXE_NULL_ARGUMENT, fname,
		"The text_data pointer passed was NULL." );
	}

	MX_DEBUG(-2,("%s invoked for socket %d, protocol_version = %ld",
		fname, mx_socket->socket_fd, protocol_version));

	MX_DEBUG(-2,("%s: sending '%s' to Blu-Ice socket %d.",
		fname, text_data, mx_socket->socket_fd ));

	bytes_sent = 0;

	text_data_length = (long) (strlen( text_data ) + 1);

	if ( protocol_version > MX_BLUICE_PROTOCOL_1 ) {
		snprintf( message_header, MX_BLUICE_MSGHDR_LENGTH,
			"%*lu%*lu",
			MX_BLUICE_MSGHDR_TEXT_LENGTH,
			(unsigned long) text_data_length,
			MX_BLUICE_MSGHDR_BINARY_LENGTH,
			(unsigned long) binary_data_length );

		/* Send the message header. */

		mx_status = mx_socket_send( mx_socket,
					&message_header,
					MX_BLUICE_MSGHDR_LENGTH );

		if ( mx_status.code != MXE_SUCCESS )
			return mx_status;

		bytes_sent += MX_BLUICE_MSGHDR_LENGTH;
	}

	/* Send the text data. */

	mx_status = mx_socket_send( mx_socket, text_data, text_data_length );

	if ( mx_status.code != MXE_SUCCESS )
		return mx_status;

	bytes_sent += text_data_length;

	/* Send the binary data (if any). */

	if ( (binary_data_length > 0) && (binary_data != NULL) ) {

		mx_status = mx_socket_send( mx_socket,
				binary_data, binary_data_length );

		if ( mx_status.code != MXE_SUCCESS )
			return mx_status;

		bytes_sent += binary_data_length;
	}

	/* If we are using Blu-Ice protocol 1.0, then we must send
	 * an additional null pad to fill out the message to 
	 * MX_BLUICE_OLD_MESSAGE_LENGTH bytes.
	 */

	if ( protocol_version == MX_BLUICE_PROTOCOL_1 ) {

		if ( bytes_sent < MX_BLUICE_OLD_MESSAGE_LENGTH ) {

			null_bytes_to_send =
				MX_BLUICE_OLD_MESSAGE_LENGTH - bytes_sent;

			memset( null_pad, 0, null_bytes_to_send );

			mx_status = mx_socket_send( mx_socket,
					null_pad, null_bytes_to_send );

			if ( mx_status.code != MXE_SUCCESS )
				return mx_status;
		}
	}

	return MX_SUCCESSFUL_RESULT;
}

/*------------------------------------------------------------------*/

mx_status_type
mxbl_receive_message( MX_SOCKET *mx_socket,
			char *data_buffer,
			long data_buffer_length,
			long *actual_data_length,
			double timeout_in_seconds,
			long protocol_version )
{
	static const char fname[] = "mxbl_receive_message()";

	char message_header[MX_BLUICE_MSGHDR_LENGTH+1];
	long num_bytes_available;
	long text_data_length, binary_data_length, total_data_length;
	size_t header_bytes_received, actual_bytes_received;
	unsigned long i, wait_ms, max_attempts;
	int num_items;
	mx_status_type mx_status;

	if ( mx_socket == (MX_SOCKET *) NULL ) {
		return mx_error( MXE_NULL_ARGUMENT, fname,
		"The MX_SOCKET pointer passed was NULL." );
	}
	if ( data_buffer == (char *) NULL ) {
		return mx_error( MXE_NULL_ARGUMENT, fname,
		"The data_buffer pointer passed was NULL." );
	}

	MX_DEBUG( 2,("%s invoked for socket %d, protocol_version = %ld",
		fname, mx_socket->socket_fd, protocol_version));

	/* Wait for a new message to come in.
	 *
	 * A negative timeout means wait forever, so we only need to loop
	 * if the timeout is >= 0.
	 */

	if ( timeout_in_seconds >= 0.0 ) {
		wait_ms = 10;
		max_attempts =
		    mx_round((1000.0 * timeout_in_seconds) / (double) wait_ms);

		for ( i = 0; i < max_attempts; i++ ) {
			mx_status = mx_socket_num_input_bytes_available(
					mx_socket, &num_bytes_available );

			if ( mx_status.code != MXE_SUCCESS )
				return mx_status;

			if ( num_bytes_available != 0 ) {
				break;		/* Exit the for() loop. */
			}

			mx_msleep( wait_ms );
		}

		if ( i >= max_attempts ) {
			return mx_error( MXE_TIMED_OUT | MXE_QUIET, fname,
				"Timed out after %g seconds waiting for "
				"socket %d to send a message.",
					timeout_in_seconds,
					mx_socket->socket_fd );
		}
	}

	/* Read in the header of the Blu-Ice message. */

	if ( protocol_version == MX_BLUICE_PROTOCOL_1 ) {

		text_data_length = MX_BLUICE_OLD_MESSAGE_LENGTH;
		binary_data_length = 0;

		header_bytes_received = 0;
	} else {
		/* This message is in Blu-Ice protocol version 2.0 
		 * or higher.
		 */

		/* Receive the message header. */

		mx_status = mx_socket_receive( mx_socket,
						message_header,
						MX_BLUICE_MSGHDR_LENGTH,
						&header_bytes_received,
						NULL, 0 );

		if ( mx_status.code != MXE_SUCCESS )
			return mx_status;

		if ( header_bytes_received != MX_BLUICE_MSGHDR_LENGTH ) {
			mx_warning( "The Blu-Ice message received from "
			"socket %d was %ld bytes long, when it should "
			"have been %d bytes long.",
				mx_socket->socket_fd,
				(long) header_bytes_received,
				MX_BLUICE_MSGHDR_LENGTH );
		}

		/* Make sure that the message header is null terminated. */

		message_header[MX_BLUICE_MSGHDR_LENGTH-1] = '\0';

		num_items = sscanf( message_header, "%ld %ld",
				&text_data_length, &binary_data_length );

		if ( num_items != 2 ) {
			return mx_error( MXE_UNPARSEABLE_STRING, fname,
			"Did not see two ASCII number is the header '%s' "
			"from Blu-Ice socket %d.",
				message_header, mx_socket->socket_fd );
		}
	}

	total_data_length = text_data_length + binary_data_length;

	MX_DEBUG(-2,("%s: text = %ld, binary = %ld, total = %ld",
		fname, text_data_length, binary_data_length,
		total_data_length));

	if ( total_data_length > data_buffer_length ) {
		mx_warning( "Truncating a %ld byte message body from "
		"Blu-Ice socket %d to %ld bytes.",
			total_data_length,
			mx_socket->socket_fd,
			data_buffer_length );

		total_data_length = data_buffer_length;
	}

	/* Now receive the message body for the message. */

	mx_status = mx_socket_receive( mx_socket,
					data_buffer,
					total_data_length,
					&actual_bytes_received,
					NULL, 0 );

	if ( actual_data_length != (long *) NULL ) {
		*actual_data_length = (long) actual_bytes_received;
	}

	if ( actual_bytes_received < data_buffer_length ) {
		data_buffer[actual_bytes_received] = '\0';
	}

	MX_DEBUG(-2,("%s: received '%s' from Blu-Ice socket %d.",
		fname, data_buffer, mx_socket->socket_fd));

	/* If the message was truncated, discard any unread input. */

	if ( total_data_length > data_buffer_length ) {
		mx_warning( "Discarding unread input from Blu-Ice socket %d.",
			mx_socket->socket_fd );

		mx_status = mx_socket_discard_unread_input( mx_socket );
	}

	return MX_SUCCESSFUL_RESULT;
}

