/*
 * Name:    mxbl_operations.c
 *
 * Purpose: Blu-Ice operations supported by MX.
 *
 * Author:  William Lavender
 *
 *--------------------------------------------------------------------------
 *
 * Copyright 2008 Illinois Institute of Technology
 *
 * See the file "LICENSE" for information on usage and redistribution
 * of this file, and for a DISCLAIMER OF ALL WARRANTIES.
 *
 */

#include <stdio.h>
#include <stdlib.h>
#include <errno.h>
#include <math.h>

#include "mx_util.h"
#include "mx_unistd.h"
#include "mx_record.h"
#include "mx_driver.h"
#include "mx_list.h"
#include "mx_socket.h"
#include "mx_callback.h"
#include "mx_process.h"
#include "mx_image.h"
#include "mx_area_detector.h"
#include "mx_scan.h"
#include "mx_scan_area_detector.h"
#include "sa_wedge.h"

#include "mx_bluice.h"
#include "d_bluice_area_detector.h"
#include "d_bluice_motor.h"
#include "d_bluice_shutter.h"
#include "f_custom.h"

#include "mx_poison.h"

#include "mxbl.h"

#define mx_atomic_increment(x) ((x)++)		/* FIXME! */

/*------------------------------------------------------------------*/

mx_status_type
mxbl_broadcast_dcss_operation_error(
	MXBL_OPERATION_ARGS *operation_args,
	long error_code,
	const char *location,
	char *format, ... ) MX_PRINTFLIKE( 4, 5 );

mx_status_type
mxbl_forward_dcss_operation_error(
	MXBL_OPERATION_ARGS *operation_args,
	mx_status_type mx_status );

mx_status_type
mxbl_send_dhs_operation_error(
	MXBL_OPERATION_ARGS *operation_args,
	long error_code,
	const char *location,
	char *format, ... ) MX_PRINTFLIKE( 4, 5 );

mx_status_type
mxbl_forward_dhs_operation_error(
	MXBL_OPERATION_ARGS *operation_args,
	mx_status_type mx_status );

mx_status_type
mxbl_start_operation( MXBL_EVENT *event,
		MXBL_OPERATION_ARGS **operation_args );

/*------------------------------------------------------------------*/

typedef mx_status_type
(MXBL_OPERATION_CALLBACK_FUNCTION)( MX_CALLBACK_MESSAGE * );

static MXBL_OPERATION_CALLBACK_FUNCTION mxbl_op_default_destructor;
static MXBL_OPERATION_CALLBACK_FUNCTION
				mxbl_op_detector_collect_image_destructor;

static MXBL_OPERATION_CALLBACK_FUNCTION mxbl_op_collectFrame;
static MXBL_OPERATION_CALLBACK_FUNCTION mxbl_op_collectRun;
static MXBL_OPERATION_CALLBACK_FUNCTION mxbl_op_collectRuns;
static MXBL_OPERATION_CALLBACK_FUNCTION mxbl_op_detector_collect_image;
static MXBL_OPERATION_CALLBACK_FUNCTION mxbl_op_detector_oscillation_ready;
static MXBL_OPERATION_CALLBACK_FUNCTION mxbl_op_detector_stop;
static MXBL_OPERATION_CALLBACK_FUNCTION mxbl_op_detector_transfer_image;
static MXBL_OPERATION_CALLBACK_FUNCTION mxbl_op_scanMotor;

static struct {
	char operation_name[40];
	MXBL_OPERATION_CALLBACK_FUNCTION *callback_function;
	MXBL_OPERATION_CALLBACK_FUNCTION *callback_destructor;
} operation_callback_list[] = {
	{"collectFrame", mxbl_op_collectFrame, mxbl_op_default_destructor},
	{"collectRun", mxbl_op_collectRun, mxbl_op_default_destructor},
	{"collectRuns", mxbl_op_collectRuns, mxbl_op_default_destructor},
	{"detector_collect_image",
		mxbl_op_detector_collect_image,
		mxbl_op_detector_collect_image_destructor},
	{"detector_oscillation_ready",
		mxbl_op_detector_oscillation_ready, mxbl_op_default_destructor},
	{"detector_stop", mxbl_op_detector_stop, mxbl_op_default_destructor},
	{"detector_transfer_image",
		mxbl_op_detector_transfer_image, mxbl_op_default_destructor},
	{"scanMotor", mxbl_op_scanMotor, mxbl_op_default_destructor},
};

static int num_operation_callbacks = sizeof( operation_callback_list )
				/ sizeof( operation_callback_list[0] );

/*------------------------------------------------------------------*/

/* FIXME: Get rid of this global variable! */

MXBL_OPERATION_ARGS *global_detector_collect_image_args = NULL;

/*------------------------------------------------------------------*/

static mx_bool_type
mxblp_string_is_null( char *string ) {
	if ( string == NULL ) {
		return TRUE;
	}
	if ( strlen(string) == 0 ) {
		return TRUE;
	}
	if ( strcmp(string, "{}") == 0 ) {
		return TRUE;
	}
	if ( strcmp(string, "NULL") == 0 ) {
		return TRUE;
	}

	return FALSE;
}

/*------------------------------------------------------------------*/

mx_status_type
mxbl_broadcast_dcss_operation_error( MXBL_OPERATION_ARGS *operation_args,
					long error_code,
					const char *location,
					char *format, ... )
{
	mx_status_type mx_status;

	va_list args;
	char mx_error_message[2500];
	char bluice_error_message[2500];

	va_start( args, format );
	vsnprintf( mx_error_message, sizeof(mx_error_message), format, args );
	va_end( args );

	mx_status = mx_error( error_code, location, mx_error_message );

	snprintf( bluice_error_message, sizeof(bluice_error_message),
		"stog_operation_completed %s %lu.%lu error %s",
			operation_args->operation_name,
			operation_args->client_number,
			operation_args->operation_number,
			mx_error_message );

	(void) mxbl_broadcast_dcss_message(
			operation_args->event->event_manager,
			bluice_error_message );

	return mx_status;
}

/*------------------------------------------------------------------*/

mx_status_type
mxbl_forward_dcss_operation_error( MXBL_OPERATION_ARGS *operation_args,
					mx_status_type mx_status )
{
	char bluice_error_message[2500];

	snprintf( bluice_error_message, sizeof(bluice_error_message),
		"stog_operation_completed %s %lu.%lu error %s",
			operation_args->operation_name,
			operation_args->client_number,
			operation_args->operation_number,
			mx_status.message );

	(void) mxbl_broadcast_dcss_message(
			operation_args->event->event_manager,
			bluice_error_message );

	return mx_status;
}

/*------------------------------------------------------------------*/

mx_status_type
mxbl_send_dhs_operation_error( MXBL_OPERATION_ARGS *operation_args,
				long error_code,
				const char *location,
				char *format, ... )
{
	mx_status_type mx_status;

	va_list args;
	char mx_error_message[2500];
	char bluice_error_message[2500];
	MX_SOCKET *mx_socket;

	va_start( args, format );
	vsnprintf( mx_error_message, sizeof(mx_error_message), format, args );
	va_end( args );

	mx_status = mx_error( error_code, location, mx_error_message );

	snprintf( bluice_error_message, sizeof(bluice_error_message),
		"stog_operation_completed %s %lu.%lu error %s",
			operation_args->operation_name,
			operation_args->client_number,
			operation_args->operation_number,
			mx_error_message );

	mx_socket = operation_args->event->mx_socket;

	(void) mxbl_send_message( mx_socket, bluice_error_message,
				NULL, 0, MX_BLUICE_PROTOCOL_2 );

	return mx_status;
}

/*------------------------------------------------------------------*/

mx_status_type
mxbl_forward_dhs_operation_error( MXBL_OPERATION_ARGS *operation_args,
					mx_status_type mx_status )
{
	char bluice_error_message[2500];
	MX_SOCKET *mx_socket;

	snprintf( bluice_error_message, sizeof(bluice_error_message),
		"stog_operation_completed %s %lu.%lu error %s",
			operation_args->operation_name,
			operation_args->client_number,
			operation_args->operation_number,
			mx_status.message );

	mx_socket = operation_args->event->mx_socket;

	(void) mxbl_send_message( mx_socket, bluice_error_message,
				NULL, 0, MX_BLUICE_PROTOCOL_2 );

	return mx_status;
}

/*------------------------------------------------------------------*/

static MX_RECORD *
mxblp_find_area_detector( MX_CALLBACK_MESSAGE *callback_message )
{
	MX_LIST_HEAD *list_head;
	MX_RECORD *record_list;
	MX_RECORD *current_record;
	MX_RECORD *ad_record;

	list_head = callback_message->list_head;

	record_list = list_head->record;

	ad_record = NULL;

	/* The _last_ area detector record in the database is the
	 * one that is returned.
	 */

	current_record = record_list->next_record;

	while( current_record != record_list ) {
		if ( (current_record->mx_superclass == MXR_DEVICE)
		  && (current_record->mx_class == MXC_AREA_DETECTOR) )
		{
			ad_record = current_record;
		}
		current_record = current_record->next_record;
	}

	return ad_record;
}

/*------------------------------------------------------------------*/

static MX_RECORD *
mxblp_find_motor( MX_CALLBACK_MESSAGE *callback_message, char *motor_name )
{
	static const char fname[] = "mxblp_find_motor()";

	MX_LIST_HEAD *list_head;
	MX_RECORD *record_list;
	MX_RECORD *current_record;
	MX_BLUICE_MOTOR *bluice_motor;

	MX_DEBUG(-2,("%s invoked for motor '%s'",
		fname, motor_name ));

	list_head = callback_message->list_head;

	record_list = list_head->record;

	current_record = record_list->next_record;

	while( current_record != record_list ) {
	    switch( current_record->mx_type ) {
	    case MXT_MTR_BLUICE_DCSS:
	    case MXT_MTR_BLUICE_DHS:
		bluice_motor = current_record->record_type_struct;

		if ( strcmp(bluice_motor->bluice_name, motor_name) == 0 ) {
			return current_record;
		}
	    }
	    current_record = current_record->next_record;
	}

	return NULL;
}

/*------------------------------------------------------------------*/

static MX_RECORD *
mxblp_find_shutter( MX_CALLBACK_MESSAGE *callback_message, char *shutter_name )
{
	static const char fname[] = "mxblp_find_shutter()";

	MX_LIST_HEAD *list_head;
	MX_RECORD *record_list;
	MX_RECORD *current_record;
	MX_BLUICE_SHUTTER *bluice_shutter;

	MX_DEBUG(-2,("%s invoked for shutter '%s'",
		fname, shutter_name ));

	list_head = callback_message->list_head;

	record_list = list_head->record;

	current_record = record_list->next_record;

	while( current_record != record_list ) {
	    switch( current_record->mx_type ) {
	    case MXT_RLY_BLUICE_DCSS_SHUTTER:
	    case MXT_RLY_BLUICE_DHS_SHUTTER:
		bluice_shutter = current_record->record_type_struct;

		if ( strcmp(bluice_shutter->bluice_name, shutter_name) == 0 ) {
			return current_record;
		}
	    }
	    current_record = current_record->next_record;
	}

	return NULL;
}

/*------------------------------------------------------------------*/

static mx_status_type
mxblp_image_write_file( MX_AREA_DETECTOR *area_detector,
			char *output_filename )
{
	static const char fname[] = "mxblp_image_write_file()";

	int os_status, saved_errno;
	mx_status_type mx_status;

	/* Does the requested file already exist? */

	os_status = access( output_filename, F_OK );

	saved_errno = errno;

	MX_DEBUG(-2,
	("%s: output_filename = '%s', os_status = %d, errno = %d",
		fname, output_filename, os_status, saved_errno));

	/* If the file does not already exist, then we should
	 * get back 'os_status == -1' and 'errno == ENOENT'.
	 */

	if ( (os_status == -1) && (saved_errno == ENOENT) ) {

		/* All is well, so continue. */
	} else {
		/* Something went wrong. */

		if ( os_status == 0 ) {
			/* The file already exists. */

			return mx_error( MXE_ALREADY_EXISTS, fname,
			"The requested output file '%s' already exists.",
		    		output_filename );
		} else {
			/* Some other sort of error occurred. */

			return mx_error( MXE_OPERATING_SYSTEM_ERROR, fname,
				"An operating system error occurred while "
				"trying to access the requested output "
				"file '%s'.  Errno = %d, error message = '%s'",
					output_filename,
					saved_errno, strerror(saved_errno) );
		}
	}

	/* FIXME: In principle, we have a race condition here,
	 * since theoretically someone could create the file
	 * specified by 'output_filename' after the call to
	 * access() returns.  However, I am not going to worry
	 * about that right now.
	 */

	/* Now we are finally ready to save the image frame. */

	mx_status = mx_image_write_file( area_detector->image_frame,
					MXT_IMAGE_FILE_SMV,
					output_filename );

	return mx_status;
}

/*------------------------------------------------------------------*/

static mx_status_type
mxblp_compute_new_dark_current_frame( MX_AREA_DETECTOR *area_detector,
					int num_dark_frames,
					MX_IMAGE_FRAME **dark_frame_array )
{
	static const char fname[] = "mxblp_compute_new_dark_current_frame()";

	mx_status_type mx_status;

	MX_DEBUG(-2,
	("%s: area_detector = %p, num_dark_frames = %d, dark_frame_array = %p",
		fname, area_detector, num_dark_frames, dark_frame_array));

	if ( num_dark_frames < 1 ) {
		MX_DEBUG(-2,("%s returning since num_dark_frames = %d",
			fname, num_dark_frames));

		return MX_SUCCESSFUL_RESULT;
	}

	if ( area_detector == (MX_AREA_DETECTOR *) NULL ) {
		return mx_error( MXE_NULL_ARGUMENT, fname,
		"The MX_AREA_DETECTOR pointer passed was NULL." );
	}
	if ( dark_frame_array == (MX_IMAGE_FRAME **) NULL ) {
		return mx_error( MXE_NULL_ARGUMENT, fname,
		"The dark current frame array pointer passed was NULL." );
	}

#if 0
	(void) mx_print_structure( stderr,
		area_detector->record, MXFF_SHOW_ALL );

	MX_DEBUG(-2,("%s: dark_current_frame = %p",
		fname, area_detector->dark_current_frame));
	MX_DEBUG(-2,("%s: framesize = (%lu,%lu)",
	    fname, area_detector->framesize[0], area_detector->framesize[1]));
	MX_DEBUG(-2,("%s: image_format = %lu",
		fname, area_detector->image_format));
	MX_DEBUG(-2,("%s: byte_order = %lu", fname, area_detector->byte_order));
	MX_DEBUG(-2,("%s: bytes_per_pixel = %f",
		fname, area_detector->bytes_per_pixel));
	MX_DEBUG(-2,("%s: header_length = %lu",
		fname, area_detector->header_length));
	MX_DEBUG(-2,("%s: bytes_per_frame = %lu",
		fname, area_detector->bytes_per_frame));
#endif

	/* Make sure that memory has already been allocated for
	 * the dark current frame, since mx_image_dezinger()
	 * will not allocate it for us.
	 */

	mx_status = mx_area_detector_setup_frame( area_detector->record,
					&(area_detector->dark_current_frame) );

	if ( mx_status.code != MXE_SUCCESS )
		return mx_status;

	/* Now compute the dezingered dark current frame. */

	mx_status = mx_image_dezinger( &(area_detector->dark_current_frame),
					num_dark_frames,
					dark_frame_array,
				fabs( area_detector->dezinger_threshold ) );

	MX_DEBUG(-2,("%s complete.", fname ));

	return mx_status;
}

/*==================================================================*/

mx_status_type
mxblh_start_operation( MXBL_EVENT *event )
{
	mx_status_type mx_status;

	mx_status = mxbl_start_operation( event, NULL );

	return mx_status;
}

mx_status_type
mxbl_start_operation( MXBL_EVENT *event,
		MXBL_OPERATION_ARGS **returned_op_args )
{
	static const char fname[] = "mxbl_start_operation()";

	MXBL_OPERATION_ARGS *operation_args;
	char operation_name[40];
	char format[20];
	int i, num_items;
	double callback_interval_in_seconds;
	unsigned long client_number, operation_number;
	mx_status_type mx_status;

	/* Figure out which operation has been requested. */

	snprintf( format, sizeof(format),
			"%%*s %%%ds %%lu.%%lu", (int) sizeof(operation_name) );

	num_items = sscanf( event->message_buffer, format,
			operation_name, &client_number, &operation_number );

	if ( num_items != 3 ) {
		return mx_error( MXE_NETWORK_IO_ERROR, fname,
		"Unparseable start operation message '%s' "
		"received by socket %d.",
			event->message_buffer, event->mx_socket->socket_fd );
	}

	MX_DEBUG(-2,("%s: Operation '%s' requested.", fname, operation_name));

	/* Look for the operation callback function. */

	for ( i = 0; i < num_operation_callbacks; i++ ) {
		if ( strcmp( operation_name,
			operation_callback_list[i].operation_name ) == 0 )
		{
			break;		/* Exit the for(i) loop. */
		}
	}

	if ( i >= num_operation_callbacks ) {
		return mx_error ( MXE_UNSUPPORTED, fname,
			"Operation '%s' is not currently supported.",
			operation_name );
	}

	operation_args = calloc( 1, sizeof(MXBL_OPERATION_ARGS) );

	if ( operation_args == NULL ) {
		return mx_error( MXE_OUT_OF_MEMORY, fname,
			"Ran out of memory trying to allocate "
			"an MXBL_OPERATION_ARGS structure." );
	}

	if ( returned_op_args != (MXBL_OPERATION_ARGS **) NULL ) {
		*returned_op_args = operation_args;
	}

	operation_args->event = event;

	strlcpy( operation_args->operation_name, operation_name,
			sizeof(operation_args->operation_name) );

	operation_args->client_number = client_number;
	operation_args->operation_number = operation_number;

	operation_args->start_message = strdup( event->message_buffer );

	if ( operation_args->start_message == NULL ) {
		return mx_error( MXE_OUT_OF_MEMORY, fname,
		"Ran out of memory trying to copy the message buffer "
		"for MXBL_EVENT %p", event );
	}

	operation_args->operation_in_progress = TRUE;

	operation_args->op_private = NULL;

	callback_interval_in_seconds = 0.1;

	mx_status = mx_function_add_callback(
				event->event_manager->record_list,
				operation_callback_list[i].callback_function,
				operation_callback_list[i].callback_destructor,
				operation_args,
				callback_interval_in_seconds,
				NULL );
	return mx_status;
}

/*==================================================================*/

/* Values for the 'state' variable below. */

#define MXBLS_CF_ERROR				(-1)
#define MXBLS_CF_PREPARING_FOR_ACQUISITION	0
#define MXBLS_CF_ACQUIRING_FRAME		1
#define MXBLS_CF_ACQUISITION_COMPLETE		2
#define MXBLS_CF_READOUT_COMPLETE		3
#define MXBLS_CF_CORRECTION_COMPLETE		4
#define MXBLS_CF_TRANSFER_COMPLETE		5
#define MXBLS_CF_IMAGE_SAVE_COMPLETE		6

typedef struct mxbl_op_collect_frame_struct {
	int dark_cache_number;
	char filename[MXU_FILENAME_LENGTH+1];
	char directory[MXU_FILENAME_LENGTH+1];
	char user_name[MXU_USERNAME_LENGTH+1];
	char motor_name[MXU_BLUICE_NAME_LENGTH+1];
	char shutter_name[MXU_BLUICE_NAME_LENGTH+1];
	double delta;
	double time;
	int mode_index;
	int flush;
	int reuse_dark;

	MX_RECORD *area_detector_record;
	MX_RECORD *motor_record;
	MX_RECORD *shutter_record;

	int state;
} MXBL_OP_COLLECT_FRAME;

static mx_status_type mxblp_start_collectFrame( MX_CALLBACK_MESSAGE *,
						MXBL_OPERATION_ARGS * );

static mx_status_type mxblp_update_collectFrame( MX_CALLBACK_MESSAGE *,
						MXBL_OPERATION_ARGS * );

static mx_status_type
mxbl_op_collectFrame( MX_CALLBACK_MESSAGE *callback_message )
{
	static const char fname[] = "mxbl_op_collectFrame()";

	MXBL_OPERATION_ARGS *operation_args;
	MX_VIRTUAL_TIMER *oneshot_timer;
	double callback_interval;
	mx_status_type mx_status;

	MX_DEBUG(-2,("%s invoked.", fname));

	if ( callback_message == (MX_CALLBACK_MESSAGE *) NULL ) {
		return mx_error( MXE_INVALID_CALLBACK, fname,
		"The MX_CALLBACK_MESSAGE pointer passed was NULL." );
	}

	operation_args = callback_message->u.function.callback_args;

	if ( operation_args == (MXBL_OPERATION_ARGS *) NULL ) {
		return mx_error( MXE_INVALID_CALLBACK, fname,
		"The callback args pointer for callback message %p is NULL.",
			callback_message );
	}

	oneshot_timer = callback_message->u.function.oneshot_timer;

	/* Has the op_collectFrame structure been initialized yet? */

	if ( operation_args->op_private == NULL ) {
		mx_status = mxblp_start_collectFrame( callback_message,
							operation_args );
	} else {
		mx_status = mxblp_update_collectFrame( callback_message,
							operation_args );
	}

	if ( mx_status.code != MXE_SUCCESS ) {
		operation_args->operation_in_progress = FALSE;
	}

	MX_DEBUG(-2,("%s: operation_in_progress = %d",
		fname, (int) operation_args->operation_in_progress));

	/* If the operation is still in progress, reschedule the callback. */

	if ( operation_args->operation_in_progress ) {
		callback_interval =
			callback_message->u.function.callback_interval;

		mx_status = mx_virtual_timer_start( oneshot_timer,
						callback_interval );
	} else {
		mx_status = mx_function_delete_callback( callback_message );
	}

	return mx_status;
}

static mx_status_type
mxblp_start_collectFrame( MX_CALLBACK_MESSAGE *callback_message,
			MXBL_OPERATION_ARGS *operation_args )
{
	static const char fname[] = "mxblp_start_collectFrame()";

	MXBL_OP_COLLECT_FRAME *op_collectFrame;
	MXBL_EVENT_MANAGER *event_manager;
	int argc;
	char **argv;
	MX_RECORD *area_detector_record;
	size_t directory_length, filename_length;
	mx_status_type mx_status;

	MX_DEBUG(-2,("%s invoked", fname));

	event_manager = operation_args->event->event_manager;

	area_detector_record = mxblp_find_area_detector( callback_message );

	if ( area_detector_record == NULL ) {
		return mx_error( MXE_NOT_FOUND, fname,
		"The MX database does not contain any area detector records.");
	}

	MX_DEBUG(-2,("%s: area detector = '%s'",
		fname, area_detector_record->name ));

	/* Setup a new op_collectFrame structure. */

	op_collectFrame = calloc( 1, sizeof(MXBL_OP_COLLECT_FRAME) );

	if ( op_collectFrame == NULL ) {
		return mx_error( MXE_OUT_OF_MEMORY, fname,
			"Ran out of memory trying to allocate an "
			"MXBL_OP_COLLECT_FRAME structure." );
	}

	operation_args->op_private = op_collectFrame;

	op_collectFrame->area_detector_record = area_detector_record;

	op_collectFrame->state = MXBLS_CF_PREPARING_FOR_ACQUISITION;

	/* Split up the start message into tokens. */

	MX_DEBUG(-2,("%s: start message = '%s'",
			fname, operation_args->start_message ));

	mx_string_split( operation_args->start_message, " ", &argc, &argv );

	if ( argc != 14 ) {
		mx_free(argv);

		return mx_error( MXE_NETWORK_IO_ERROR, fname,
			"The start_operation message from the Blu-Ice server "
			"had an incorrect number of arguments.  "
			"The correct number of arguments is 14." );
	}

	op_collectFrame->dark_cache_number = atoi( argv[3] );

	if ( !mxblp_string_is_null(argv[4]) ) {
		strlcpy( op_collectFrame->filename, argv[4],
			sizeof(op_collectFrame->filename) );
	}
	if ( !mxblp_string_is_null(argv[5]) ) {
		strlcpy( op_collectFrame->directory, argv[5],
			sizeof(op_collectFrame->directory) );
	}
	if ( !mxblp_string_is_null(argv[6]) ) {
		strlcpy( op_collectFrame->user_name, argv[6],
			sizeof(op_collectFrame->user_name) );
	}
	if ( !mxblp_string_is_null(argv[7]) ) {
		strlcpy( op_collectFrame->motor_name, argv[7],
			sizeof(op_collectFrame->motor_name) );
	}
	if ( !mxblp_string_is_null(argv[8]) ) {
		strlcpy( op_collectFrame->shutter_name, argv[8],
			sizeof(op_collectFrame->shutter_name) );
	}

	op_collectFrame->delta      = atof( argv[9] );
	op_collectFrame->time       = atof( argv[10] );
	op_collectFrame->mode_index = atoi( argv[11] );
	op_collectFrame->flush      = atoi( argv[12] );
	op_collectFrame->reuse_dark = atoi( argv[13] );

#if 1
	MX_DEBUG(-2,("%s:  dark_cache_number = %d",
		fname, op_collectFrame->dark_cache_number));
	MX_DEBUG(-2,("%s:  filename = '%s'",
		fname, op_collectFrame->filename));
	MX_DEBUG(-2,("%s:  directory = '%s'",
		fname, op_collectFrame->directory));
	MX_DEBUG(-2,("%s:  user_name = '%s'",
		fname, op_collectFrame->user_name));
	MX_DEBUG(-2,("%s:  motor_name = '%s'",
		fname, op_collectFrame->motor_name));
	MX_DEBUG(-2,("%s:  shutter_name = '%s'",
		fname, op_collectFrame->shutter_name));
	MX_DEBUG(-2,("%s:  delta = %lf",
		fname, op_collectFrame->delta));
	MX_DEBUG(-2,("%s:  time = %lf",
		fname, op_collectFrame->time));
	MX_DEBUG(-2,("%s:  mode_index = %d",
		fname, op_collectFrame->mode_index));
	MX_DEBUG(-2,("%s:  flush = %d",
		fname, op_collectFrame->flush));
	MX_DEBUG(-2,("%s:  reuse_dark = %d",
		fname, op_collectFrame->reuse_dark));
#endif
	/* If requested, look for the specified motor and shutter records. */

	op_collectFrame->motor_record = mxblp_find_motor( callback_message,
						op_collectFrame->motor_name );

	if ( strlen(op_collectFrame->motor_name) == 0 ) {
		/* Do nothing if no motor name was specified. */
	} else
	if ( op_collectFrame->motor_record == NULL ) {
		(void) mx_error( MXE_NOT_FOUND, fname,
		"Blu-Ice motor '%s' was not found in the MX database.",
			op_collectFrame->motor_name );
	} else {
		MX_DEBUG(-2,("%s: Blu-Ice motor '%s' == MX motor '%s'",
		fname, op_collectFrame->motor_name,
		op_collectFrame->motor_record->name ));
	}

	op_collectFrame->shutter_record = mxblp_find_shutter( callback_message,
						op_collectFrame->shutter_name );

	if ( strlen(op_collectFrame->shutter_name) == 0 ) {
		/* Do nothing if no shutter name was specified. */
	} else
	if ( op_collectFrame->shutter_record == NULL ) {
		(void) mx_error( MXE_NOT_FOUND, fname,
		"Blu-Ice shutter '%s' was not found in the MX database.",
			op_collectFrame->shutter_name );
	} else {
		MX_DEBUG(-2,("%s: Blu-Ice shutter '%s' == MX shutter '%s'",
		fname, op_collectFrame->shutter_name,
		op_collectFrame->shutter_record->name ));
	}

	mx_free( operation_args->start_message );
	mx_free( argv );

	/* Verify that non-empty values have been specified for the
	 * directory and file names.
	 */

	directory_length = strlen( op_collectFrame->directory );
	filename_length = strlen( op_collectFrame->filename );

	if ( (directory_length == 0) && (filename_length == 0) ) {

		return mxbl_broadcast_dcss_operation_error( operation_args,
			MXE_INITIALIZATION_ERROR, fname,
			"No directory or filename have been specified "
			"for area_detector '%s'.", area_detector_record->name );
	} else
	if ( directory_length == 0 ) {
		return mxbl_broadcast_dcss_operation_error( operation_args,
			MXE_INITIALIZATION_ERROR, fname,
			"No directory name has been specified "
			"for area_detector '%s'.", area_detector_record->name );
	} else
	if ( filename_length == 0 ) {
		return mxbl_broadcast_dcss_operation_error( operation_args,
			MXE_INITIALIZATION_ERROR, fname,
			"No filename has been specified "
			"for area_detector '%s'.", area_detector_record->name );
	}

	/* Select one-shot mode for the area detector with the 
	 * requested exposure time.
	 */

	MX_DEBUG(-2,
    ("%s: Setting area detector '%s' to one-shot mode with exposure time = %f",
    		fname, area_detector_record->name, op_collectFrame->time));

	mx_status = mx_area_detector_set_one_shot_mode( area_detector_record,
							op_collectFrame->time );
	if ( mx_status.code != MXE_SUCCESS ) {
		return mxbl_forward_dcss_operation_error(
					operation_args, mx_status );
	}

	MX_DEBUG(-2,("%s: Starting area detector '%s'",
		fname, area_detector_record->name));

	/* Start the exposure. */

	mx_status = mx_area_detector_start( area_detector_record );

	if ( mx_status.code != MXE_SUCCESS ) {
		return mxbl_forward_dcss_operation_error(
					operation_args, mx_status );
	}

	op_collectFrame->state = MXBLS_CF_ACQUIRING_FRAME;

	MX_DEBUG(-2,("%s complete.", fname));

	return MX_SUCCESSFUL_RESULT;
}

static mx_status_type
mxblp_update_collectFrame( MX_CALLBACK_MESSAGE *callback_message,
			MXBL_OPERATION_ARGS *operation_args )
{
	static const char fname[] = "mxblp_update_collectFrame()";

	MXBL_OP_COLLECT_FRAME *op_collectFrame;
	MXBL_EVENT_MANAGER *event_manager;
	MX_RECORD *area_detector_record;
	MX_AREA_DETECTOR *area_detector;
	char message[200];
	char output_filename[MXU_FILENAME_LENGTH+1];
	long last_frame_number, total_num_frames;
	unsigned long ad_status;
	mx_status_type mx_status;

	op_collectFrame = operation_args->op_private;

	event_manager = operation_args->event->event_manager;

	area_detector_record = op_collectFrame->area_detector_record;

	area_detector = area_detector_record->record_class_struct;

	MX_DEBUG(-2,("%s invoked for area detector '%s'",
		fname, area_detector_record->name));

	switch( op_collectFrame->state ) {
	case MXBLS_CF_ACQUIRING_FRAME:
		MX_DEBUG(-2,("%s: state = MXBLS_CF_ACQUIRING_FRAME", fname));

		mx_status = mx_area_detector_get_extended_status(
							area_detector_record,
							&last_frame_number,
							&total_num_frames,
							&ad_status );

		if ( mx_status.code != MXE_SUCCESS ) {
			op_collectFrame->state = MXBLS_CF_ERROR;

			return mxbl_forward_dcss_operation_error(
					operation_args, mx_status );
		}

		MX_DEBUG(-2,
    ("%s: last_frame_number = %ld, total_num_frames = %ld, ad_status = %#lx",
    			fname, last_frame_number, total_num_frames, ad_status));

		/* If the area detector is busy, we just return now. */

		if ( ad_status & MXSF_AD_IS_BUSY ) {
			return MX_SUCCESSFUL_RESULT;
		}

		/* Did the area detector abort with an error? */

		if ( ad_status & MXSF_AD_ERROR ) {
			op_collectFrame->state = MXBLS_CF_ERROR;

			return mxbl_broadcast_dcss_operation_error(
				operation_args, MXE_INTERRUPTED, fname,
				"Area detector '%s' aborted frame acquisition "
				"with an error.  detector status flags = %#lx",
					area_detector_record->name,
					ad_status );
		}

		/* If we get here, the image acquisition is complete.
		 * Change the state accordingly and then return.
		 */

		op_collectFrame->state = MXBLS_CF_ACQUISITION_COMPLETE;
		break;

	case MXBLS_CF_ACQUISITION_COMPLETE:
		MX_DEBUG(-2,
		("%s: state = MXBLS_CF_ACQUISITION_COMPLETE", fname));

		/* Now read out the image frame from the detector hardware.
		 * We always use frame number 0, since this server is not
		 * written to handle multiframe sequences.
		 */

		mx_status = mx_area_detector_readout_frame(
					area_detector_record, 0 );
							
		if ( mx_status.code != MXE_SUCCESS ) {
			op_collectFrame->state = MXBLS_CF_ERROR;

			return mxbl_forward_dcss_operation_error(
					operation_args, mx_status );
		}

		op_collectFrame->state = MXBLS_CF_READOUT_COMPLETE;
		break;

	case MXBLS_CF_READOUT_COMPLETE:
		MX_DEBUG(-2,("%s: state = MXBLS_CF_READOUT_COMPLETE", fname));

		/* The next step is to correct the frame. */

		mx_status = mx_area_detector_correct_frame(
						area_detector_record );

		if ( mx_status.code != MXE_SUCCESS ) {
			op_collectFrame->state = MXBLS_CF_ERROR;

			return mxbl_forward_dcss_operation_error(
					operation_args, mx_status );
		}

		op_collectFrame->state = MXBLS_CF_CORRECTION_COMPLETE;
		break;

	case MXBLS_CF_CORRECTION_COMPLETE:
		MX_DEBUG(-2,("%s: state = MXBLS_CF_CORRECTION_COMPLETE",fname));

		/* We now transfer the frame to a local frame buffer. */

		mx_status = mx_area_detector_transfer_frame(
						area_detector_record,
						MXFT_AD_IMAGE_FRAME,
						area_detector->image_frame );

		if ( mx_status.code != MXE_SUCCESS ) {
			op_collectFrame->state = MXBLS_CF_ERROR;

			return mxbl_forward_dcss_operation_error(
					operation_args, mx_status );
		}

		op_collectFrame->state = MXBLS_CF_TRANSFER_COMPLETE;
		break;

	case MXBLS_CF_TRANSFER_COMPLETE:
		MX_DEBUG(-2,("%s: state = MXBLS_CF_TRANSFER_COMPLETE",fname));

		/* Construct the name of the file to save the image to. */

		snprintf( output_filename, sizeof(output_filename),
			"%s/%s.img", op_collectFrame->directory,
			op_collectFrame->filename );

		/* Write out the file. */

		mx_status = mxblp_image_write_file( area_detector,
							output_filename );

		if ( mx_status.code != MXE_SUCCESS ) {
			op_collectFrame->state = MXBLS_CF_ERROR;

			return mxbl_forward_dcss_operation_error(
					operation_args, mx_status );
		}

		op_collectFrame->state = MXBLS_CF_IMAGE_SAVE_COMPLETE;

		/* Send out a message updating the contents of the
		 * lastImageCollected string.
		 */

		snprintf( message, sizeof(message),
		    "stog_set_string_completed lastImageCollected normal %s",
		    output_filename );

		(void) mxbl_broadcast_dcss_message( event_manager, message );
		break;

	case MXBLS_CF_IMAGE_SAVE_COMPLETE:
		MX_DEBUG(-2,("%s: state = MXBLS_CF_IMAGE_SAVE_COMPLETE",fname));

		/* The area detector finished frame acquisition successfully. */

		operation_args->operation_in_progress = FALSE;

		snprintf( message, sizeof(message),
			"stog_operation_completed collectFrame %lu.%lu normal "
			"Area detector '%s' successfully acquired a frame.",
				operation_args->client_number,
				operation_args->operation_number,
				area_detector_record->name );

		(void) mxbl_broadcast_dcss_message( event_manager, message );

		break;
	}

	return MX_SUCCESSFUL_RESULT;
}

/*==================================================================*/

/* Values for the 'state' variable for the 'collectRun' operation. */

#define MXBLS_CR_ERROR			(-1)
#define MXBLS_CR_PREPARING_FOR_RUN	0
#define MXBLS_CR_RUN_IN_PROGRESS	1
#define MXBLS_CR_RUN_COMPLETE		2

typedef struct {
	MX_RECORD *scan_record;
	MX_THREAD *scan_thread;
	mx_bool_type is_child;
	unsigned long run_number;
	MX_CALLBACK_MESSAGE *callback_message;

	int state;
} MXBL_OP_COLLECT_RUN;

static mx_status_type mxblp_start_collectRun( MX_CALLBACK_MESSAGE *,
						MXBL_OPERATION_ARGS * );

static mx_status_type mxblp_update_collectRun( MX_CALLBACK_MESSAGE *,
						MXBL_OPERATION_ARGS * );

static mx_status_type mxblp_collectRun_thread( MX_THREAD *thread, void *args );

/*----*/

static mx_status_type mxbl_collectRun_df_open( MX_DATAFILE * );
static mx_status_type mxbl_collectRun_df_close( MX_DATAFILE * );
static mx_status_type mxbl_collectRun_df_add_measurement_to_datafile(
						MX_DATAFILE * );

MX_DATAFILE_FUNCTION_LIST mxbl_collectRun_df_list = {
	mxbl_collectRun_df_open,
	mxbl_collectRun_df_close,
	NULL,
	NULL,
	NULL,
	mxbl_collectRun_df_add_measurement_to_datafile,
	NULL
};

/*----*/

static mx_status_type
mxbl_op_collectRun( MX_CALLBACK_MESSAGE *callback_message )
{
	static const char fname[] = "mxbl_op_collectRun()";

	MXBL_OPERATION_ARGS *operation_args;
	MXBL_OP_COLLECT_RUN *op_collectRun;
	MX_VIRTUAL_TIMER *oneshot_timer;
	double callback_interval;
	mx_status_type mx_status;

	MX_DEBUG(-2,("%s invoked.", fname));

	if ( callback_message == (MX_CALLBACK_MESSAGE *) NULL ) {
		return mx_error( MXE_INVALID_CALLBACK, fname,
		"The MX_CALLBACK_MESSAGE pointer passed was NULL." );
	}

	operation_args = callback_message->u.function.callback_args;

	if ( operation_args == (MXBL_OPERATION_ARGS *) NULL ) {
		return mx_error( MXE_INVALID_CALLBACK, fname,
		"The callback args pointer for callback message %p is NULL.",
			callback_message );
	}

	oneshot_timer = callback_message->u.function.oneshot_timer;

	/* Has the op_collectRun structure been initialized yet? */

	if ( operation_args->op_private == NULL ) {
		mx_status = mxblp_start_collectRun( callback_message,
							operation_args );
	} else {
		mx_status = mxblp_update_collectRun( callback_message,
							operation_args );
	}

	if ( mx_status.code != MXE_SUCCESS ) {
		operation_args->operation_in_progress = FALSE;
	}

	MX_DEBUG(-2,("%s: operation_in_progress = %d",
		fname, (int) operation_args->operation_in_progress));

	/* If the operation is still in progress, reschedule the callback. */

	if ( operation_args->operation_in_progress ) {
		callback_interval =
			callback_message->u.function.callback_interval;

		mx_status = mx_virtual_timer_start( oneshot_timer,
						callback_interval );
	} else {
		op_collectRun = operation_args->op_private;

		/* If this was a child operation, then the parent
		 * operation will delete our callback.
		 */

		if ( op_collectRun->is_child == FALSE ) {
			mx_status = mx_function_delete_callback(
						callback_message );
		}
	}

	return mx_status;
}

static mx_status_type
mxblp_start_collectRun( MX_CALLBACK_MESSAGE *callback_message,
			MXBL_OPERATION_ARGS *operation_args )
{
	static const char fname[] = "mxblp_start_collectRun()";

	MXBL_OP_COLLECT_RUN *op_collectRun;
	MXBL_EVENT_MANAGER *event_manager;
	MX_RECORD *record_list;
	int argc;
	char **argv;
	char scan_name[MXU_RECORD_NAME_LENGTH+1];
	MX_RECORD *scan_record;
	MX_SCAN *scan;
	char message[100];
	mx_status_type mx_status;

	MX_DEBUG(-2,("%s invoked", fname));

	event_manager = operation_args->event->event_manager;

	record_list = event_manager->record_list;

	/* Setup a new op_collectRun structure. */

	op_collectRun = calloc( 1, sizeof(MXBL_OP_COLLECT_RUN) );

	if ( op_collectRun == NULL ) {
		return mx_error( MXE_OUT_OF_MEMORY, fname,
			"Ran out of memory trying to allocate an "
			"MXBL_OP_COLLECT_RUN structure." );
	}

	operation_args->op_private = op_collectRun;

	op_collectRun->state = MXBLS_CR_PREPARING_FOR_RUN;

	op_collectRun->callback_message = callback_message;

	/* Split up the start message into tokens. */

	MX_DEBUG(-2,("%s: start message = '%s'",
			fname, operation_args->start_message ));

	mx_string_split( operation_args->start_message, " ", &argc, &argv );

	if ( argc != 5 ) {
		mx_free(argv);

		return mx_error( MXE_NETWORK_IO_ERROR, fname,
			"The start_operation message from the Blu-Ice server "
			"had an incorrect number of arguments (%d).  "
			"The correct number of arguments is 5.", argc );
	}

	op_collectRun->run_number = atol( argv[3] );

	/* If the username is 'self', then this operation was started
	 * internally by another operation such as 'collectRuns'.
	 */

	if ( strcmp( argv[4], "self" ) == 0 ) {
		op_collectRun->is_child = TRUE;
	} else {
		op_collectRun->is_child = FALSE;
	}

	mx_free( argv );
	mx_free( operation_args->start_message );

	snprintf( scan_name, sizeof(scan_name),
		"run%lu", op_collectRun->run_number );

	scan_record = mx_get_record( record_list, scan_name );

	if ( scan_record == (MX_RECORD *) NULL ) {
		return mx_error( MXE_NOT_FOUND, fname,
		"Scan record '%s' was not found in the MX database.",
			scan_name );
	}

	MX_DEBUG(-2,("%s: scan '%s' found.", fname, scan_record->name));

	/* Connect the Blu-Ice operation to the scan. */

	scan = scan_record->record_superclass_struct;

	scan->application_ptr = operation_args;

	/* Suppress the standard MX scan progress display. */

	scan->scan_flags |= MXF_SCAN_SUPPRESS_PROGRESS_DISPLAY;

	/*--- Redirect the scan output so that it can be sent to Blu-Ice. ---*/

	/* FIXME: At the moment, this is a memory leak. */

	mx_status = mx_scan_set_custom_datafile_handler( scan_record,
						operation_args,
						&mxbl_collectRun_df_list );

	if ( mx_status.code != MXE_SUCCESS )
		return mx_status;

	/*--- Tell the client that the run is starting. ---*/

	snprintf( message, sizeof(message),
		"collecting {Started data collection at run %lu.}",
		op_collectRun->run_number );

	mx_status = mxbl_set_collect_msg( event_manager, message );

	if ( mx_status.code != MXE_SUCCESS )
		return mx_status;

	/*-------------------------------------------------------------------*/

	/* At this point, the scan record is ready to be executed.
	 * We run the scan in a separate thread so that the DCSS
	 * process will not be blocked.
	 */

	MX_DEBUG(-2,("%s: *** Start run thread here ***", fname));

	op_collectRun->scan_record = scan_record;

	mx_status = mx_thread_create( &(op_collectRun->scan_thread),
			mxblp_collectRun_thread, operation_args );

	MX_DEBUG(-2,("%s complete.", fname));

	return mx_status;
}

static mx_status_type
mxblp_update_collectRun( MX_CALLBACK_MESSAGE *callback_message,
			MXBL_OPERATION_ARGS *operation_args )
{
	static const char fname[] = "mxblp_update_collectRun()";

	MXBL_OP_COLLECT_RUN *op_collectRun;
	MXBL_EVENT *event;
	char message[200];

	op_collectRun = operation_args->op_private;

	event = operation_args->event;

	MX_DEBUG(-2,("%s invoked.", fname));

	switch( op_collectRun->state ) {
	case MXBLS_CR_PREPARING_FOR_RUN:
		MX_DEBUG(-2,("%s: state = MXBLS_CR_PREPARE_FOR_RUN", fname));

		/* The scan has not yet started, so we do nothing here. */
		break;

	case MXBLS_CR_RUN_IN_PROGRESS:
		MX_DEBUG(-2,("%s: state = MXBLS_CR_RUN_IN_PROGRESS", fname));

		/* The scan has not yet completed, so we do nothing here. */
		break;

	case MXBLS_CR_RUN_COMPLETE:
		MX_DEBUG(-2,("%s: state = MXBLS_CR_RUN_COMPLETE", fname));

		/* the area detector run has finished successfully. */

		snprintf( message, sizeof(message),
			"inactive {%lu is complete.}",
			op_collectRun->run_number );

		operation_args->operation_in_progress = FALSE;

		(void) mxbl_broadcast_dcss_message( event->event_manager,
							message );

		snprintf( message, sizeof(message),
			"stog_operation_completed collectRun %lu.%lu normal "
			"{Run %lu is complete.}",
				operation_args->client_number,
				operation_args->operation_number,
				op_collectRun->run_number );

		(void) mxbl_broadcast_dcss_message( event->event_manager,
							message );
		break;
	}

	return MX_SUCCESSFUL_RESULT;
}

/*---*/

static mx_status_type
mxblp_collectRun_thread( MX_THREAD *thread, void *args )
{
	static const char fname[] = "mxblp_collectRun_thread()";

	MXBL_OPERATION_ARGS *operation_args;
	MXBL_OP_COLLECT_RUN *op_collectRun;
	MX_RECORD *scan_record;
	mx_status_type mx_status;

	operation_args = args;

	if ( operation_args == (MXBL_OPERATION_ARGS *) NULL ) {
		return mx_error( MXE_NULL_ARGUMENT, fname,
		"The thread arguments pointer passed was NULL." );
	}

	op_collectRun = operation_args->op_private;

	if ( op_collectRun == (MXBL_OP_COLLECT_RUN *) NULL ) {
		return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
		"The op_private pointer for this operation is NULL." );
	}

	scan_record = op_collectRun->scan_record;

	if ( scan_record == (MX_RECORD *) NULL ) {
		return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
		"The scan_record pointer for this operation is NULL." );
	}

	MX_DEBUG(-2,("%s invoked for scan record '%s'.",
		fname, scan_record->name ));

	/* Run the scan. */

	op_collectRun->state = MXBLS_CR_RUN_IN_PROGRESS;

	mx_status = mx_perform_scan( scan_record );

	MX_DEBUG(-2,("%s: scan '%s' completed with status = %ld",
		fname, scan_record->name, mx_status.code ));

	if ( mx_status.code == MXE_SUCCESS ) {
		op_collectRun->state = MXBLS_CR_RUN_COMPLETE;
	} else {
		op_collectRun->state = MXBLS_CR_ERROR;
	}

	return MX_SUCCESSFUL_RESULT;
}

/*----*/

static mx_status_type
mxbl_initialize_datafile_naming( MX_SCAN *scan )
{
	static const char fname[] = "mxbl_initialize_datafile_naming()";

	MX_RECORD *ad_record;
	MX_AREA_DETECTOR *ad;
	char *ptr;
	mx_status_type mx_status;

	MX_DEBUG(-2,("%s invoked.", fname));

	if ( scan == (MX_SCAN *) NULL ) {
		return mx_error( MXE_NULL_ARGUMENT, fname,
		"The MX_SCAN pointer passed was NULL." );
	}

	ad_record = scan->input_device_array[0];

	if ( ad_record == (MX_RECORD *) NULL ) {
		return mx_error( MXE_INITIALIZATION_ERROR, fname,
		"No input devices are specified for scan '%s'.",
			scan->record->name );
	}

	ad = ad_record->record_class_struct;

	if ( ad == (MX_AREA_DETECTOR *) NULL ) {
		return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
		"The MX_AREA_DETECTOR pointer for area detector '%s' is NULL.",
			ad_record->name );
	}

	/* The datafile naming for collectRun is almost the same as
	 * the default, so we set it to the default and then modify it.
	 */

	mx_status =
	    mxs_area_detector_scan_default_initialize_datafile_naming( scan );

	if ( mx_status.code != MXE_SUCCESS )
		return mx_status;

	/* The default pattern contains a sequence of # characters.
	 * Overwrite the first of those characters with a null byte.
	 */

	ptr = strchr( ad->datafile_pattern, '#' );

	if ( ptr != NULL ) {
		*ptr = '\0';
	}

	return MX_SUCCESSFUL_RESULT;
}

static mx_status_type
mxbl_construct_next_datafile_name( MX_SCAN *scan )
{
	static const char fname[] = "mxbl_construct_next_datafile_name()";

	MX_AREA_DETECTOR_SCAN *ad_scan;
	MX_RECORD *ad_record;
	MX_AREA_DETECTOR *ad;
	MXBL_OPERATION_ARGS *operation_args;
	MXBL_OP_COLLECT_RUN *op_collectRun;

	MX_DEBUG(-2,("%s invoked.", fname));

	if ( scan == (MX_SCAN *) NULL ) {
		return mx_error( MXE_NULL_ARGUMENT, fname,
		"The MX_SCAN pointer passed was NULL." );
	}

	ad_scan = scan->record->record_class_struct;

	if ( ad_scan == (MX_AREA_DETECTOR_SCAN *) NULL ) {
		return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
		"The MX_AREA_DETECTOR_SCAN pointer for scan '%s' is NULL.",
			scan->record->name );
	}

	ad_record = scan->input_device_array[0];

	if ( ad_record == (MX_RECORD *) NULL ) {
		return mx_error( MXE_INITIALIZATION_ERROR, fname,
		"No input devices are specified for scan '%s'.",
			scan->record->name );
	}

	ad = ad_record->record_class_struct;

	if ( ad == (MX_AREA_DETECTOR *) NULL ) {
		return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
		"The MX_AREA_DETECTOR pointer for area detector '%s' is NULL.",
			ad_record->name );
	}

	operation_args = scan->application_ptr;

	if ( operation_args == (MXBL_OPERATION_ARGS *) NULL ) {
		return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
		"The MXBL_OPERATION_ARGS pointer for scan '%s' is NULL.",
			scan->record->name );
	}

	op_collectRun = operation_args->op_private;

	if ( op_collectRun == (MXBL_OP_COLLECT_RUN *) NULL ) {
		return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
		"The MXBL_OP_COLLECT_RUN pointer for scan '%s' is NULL.",
			scan->record->name );
	}

	snprintf( ad->datafile_name, sizeof(ad->datafile_name),
	"%s_%lu_E%lu_%03lu.img",
		ad->datafile_pattern,
		op_collectRun->run_number,
		ad_scan->current_energy_number + 1,
		ad_scan->current_frame_number + 1 );

	return MX_SUCCESSFUL_RESULT;
}

/*----*/

static mx_status_type
mxbl_collectRun_df_open( MX_DATAFILE *datafile )
{
	static const char fname[] = "mxbl_collectRun_df_open()";

	MX_SCAN *scan;
	MX_RECORD *scan_record;
	mx_status_type mx_status;

	MX_DEBUG(-2,("%s invoked.", fname));

	if ( datafile == (MX_DATAFILE *) NULL ) {
		return mx_error( MXE_NULL_ARGUMENT, fname,
		"The MX_DATAFILE pointer passed was NULL." );
	}

	scan = datafile->scan;

	if ( scan == (MX_SCAN *) NULL ) {
		return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
		"The MX_SCAN pointer for MX_DATAFILE %p is NULL.",
			datafile );
	}

	scan_record = scan->record;

	if ( scan_record == (MX_RECORD *) NULL ) {
		return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
		"The scan record pointer for MX_DATAFILE %p is NULL.",
			datafile );
	}

	mx_status = mx_area_detector_scan_setup_datafile_naming( scan_record,
					mxbl_initialize_datafile_naming,
					mxbl_construct_next_datafile_name );

	return mx_status;
}

static mx_status_type
mxbl_collectRun_df_close( MX_DATAFILE *datafile )
{
	static const char fname[] = "mxbl_collectRun_df_close()";

	MX_DEBUG(-2,("%s invoked.", fname));

	/* FIXME: We currently do nothing here. */

	return MX_SUCCESSFUL_RESULT;
}

static mx_status_type
mxbl_collectRun_df_add_measurement_to_datafile( MX_DATAFILE *datafile )
{
	static const char fname[] =
		"mxbl_collectRun_df_add_measurement_to_datafile()";

	MX_DATAFILE_CUSTOM *custom_datafile;
	MXBL_OPERATION_ARGS *operation_args;
	MXBL_OP_COLLECT_RUN *op_collectRun;
	MXBL_EVENT *event;
	MX_SCAN *scan;
	MX_AREA_DETECTOR_SCAN *ad_scan;
	MX_WEDGE_SCAN *wedge_scan;
	MX_RECORD *motor_record;
	MX_BLUICE_MOTOR *bluice_motor;
	double actual_position;
	char message[100];
	mx_status_type mx_status;

	MX_DEBUG(-2,("%s invoked.", fname));

	custom_datafile = datafile->datafile_type_struct;

	operation_args = custom_datafile->custom_args;

	op_collectRun = operation_args->op_private;

	event = operation_args->event;

	scan = datafile->scan;

	ad_scan = scan->record->record_class_struct;

	wedge_scan = scan->record->record_type_struct;

#if 1
	{
		long energy_num, frame_num;
		double position;

		energy_num = ad_scan->current_energy_number;
		frame_num = ad_scan->current_frame_number;

		position = ad_scan->start_position[0]
				+ ad_scan->step_size[0] * frame_num;

		MX_DEBUG(-2,
("%s: run %lu, frame %ld, wedge = %ld, energy[%ld] = %f, position[%ld] = %f",
		fname, op_collectRun->run_number,
		ad_scan->current_frame_number,
		wedge_scan->current_wedge_number,
		energy_num, ad_scan->energy_array[energy_num],
		frame_num, position ));
	}
#endif

	/* Tell the client which frame we are on. */

	snprintf( message, sizeof(message), "collecting {%lu frame %lu}",
		op_collectRun->run_number, ad_scan->current_frame_number );

	mx_status = mxbl_set_collect_msg( event->event_manager, message );

	if ( mx_status.code != MXE_SUCCESS )
		return mx_status;

	/* Tell the client the current position of the motor. */

	motor_record = scan->motor_record_array[0];

	bluice_motor = motor_record->record_type_struct;

	mx_status = mx_motor_get_position( motor_record, &actual_position );

	if ( mx_status.code != MXE_SUCCESS )
		return mx_status;

#if 0
	snprintf( message, sizeof(message),
		"stog_update_motor_position %s %f normal",
		bluice_motor->bluice_name, actual_position );
#else
	snprintf( message, sizeof(message),
		"stog_motor_move_completed %s %f normal",
		bluice_motor->bluice_name, actual_position );
#endif

	mx_status = mxbl_broadcast_dcss_message(event->event_manager, message);

	return mx_status;
}

/*==================================================================*/

/* Values for the 'state' variable for the 'collectRuns' operation. */

#define MXBLS_CRS_ERROR			(-1)
#define MXBLS_CRS_PREPARING_FOR_RUNS	0
#define MXBLS_CRS_RUNS_ARE_ACTIVE	1
#define MXBLS_CRS_RUNS_ARE_COMPLETE	2

typedef struct mxbl_op_collect_runs_struct {
	unsigned long current_run_number;
	unsigned long total_runs;

	MXBL_OPERATION_ARGS *child_operation_args;

	int state;
} MXBL_OP_COLLECT_RUNS;

static mx_status_type mxblp_start_collectRuns( MX_CALLBACK_MESSAGE *,
						MXBL_OPERATION_ARGS * );

static mx_status_type mxblp_update_collectRuns( MX_CALLBACK_MESSAGE *,
						MXBL_OPERATION_ARGS * );

static mx_status_type
mxbl_op_collectRuns( MX_CALLBACK_MESSAGE *callback_message )
{
	static const char fname[] = "mxbl_op_collectRuns()";

	MXBL_OPERATION_ARGS *operation_args;
	MX_VIRTUAL_TIMER *oneshot_timer;
	double callback_interval;
	mx_status_type mx_status;

	MX_DEBUG(-2,("%s invoked.", fname));

	if ( callback_message == (MX_CALLBACK_MESSAGE *) NULL ) {
		return mx_error( MXE_INVALID_CALLBACK, fname,
		"The MX_CALLBACK_MESSAGE pointer passed was NULL." );
	}

	operation_args = callback_message->u.function.callback_args;

	if ( operation_args == (MXBL_OPERATION_ARGS *) NULL ) {
		return mx_error( MXE_INVALID_CALLBACK, fname,
		"The callback args pointer for callback message %p is NULL.",
			callback_message );
	}

	oneshot_timer = callback_message->u.function.oneshot_timer;

	/* Has the op_collectRuns structure been initialized yet? */

	if ( operation_args->op_private == NULL ) {
		mx_status = mxblp_start_collectRuns( callback_message,
							operation_args );
	} else {
		mx_status = mxblp_update_collectRuns( callback_message,
							operation_args );
	}

	if ( mx_status.code != MXE_SUCCESS ) {
		operation_args->operation_in_progress = FALSE;
	}

	MX_DEBUG(-2,("%s: operation_in_progress = %d",
		fname, (int) operation_args->operation_in_progress));

	/* If the operation is still in progress, reschedule the callback. */

	if ( operation_args->operation_in_progress ) {
		callback_interval =
			callback_message->u.function.callback_interval;

		mx_status = mx_virtual_timer_start( oneshot_timer,
						callback_interval );
	} else {
		mx_status = mx_function_delete_callback( callback_message );
	}

	return mx_status;
}

static mx_status_type
mxblp_start_collectRuns( MX_CALLBACK_MESSAGE *callback_message,
			MXBL_OPERATION_ARGS *operation_args )
{
	static const char fname[] = "mxblp_start_collectRuns()";

	MXBL_OP_COLLECT_RUNS *op_collectRuns;
	MXBL_EVENT_MANAGER *event_manager;
	MX_RECORD *record_list, *current_record;
	int argc;
	char **argv;
	char start_message[80];
	mx_status_type mx_status;

	MX_DEBUG(-2,("%s invoked", fname));

	event_manager = operation_args->event->event_manager;

	record_list = event_manager->record_list;

	/* Setup a new op_collectRuns structure. */

	op_collectRuns = calloc( 1, sizeof(MXBL_OP_COLLECT_RUNS) );

	if ( op_collectRuns == NULL ) {
		return mx_error( MXE_OUT_OF_MEMORY, fname,
			"Ran out of memory trying to allocate an "
			"MXBL_OP_COLLECT_RUNS structure." );
	}

	operation_args->op_private = op_collectRuns;

	op_collectRuns->state = MXBLS_CRS_PREPARING_FOR_RUNS;

	/* Split up the start message into tokens. */

	MX_DEBUG(-2,("%s: start message = '%s'",
			fname, operation_args->start_message ));

	mx_string_split( operation_args->start_message, " ", &argc, &argv );

	if ( argc != 4 ) {
		mx_free(argv);

		return mx_error( MXE_NETWORK_IO_ERROR, fname,
			"The start_operation message from the Blu-Ice server "
			"had an incorrect number of arguments (%d).  "
			"The correct number of arguments is 4.", argc );
	}

	/* How many runs are currently in the MX database? */

	op_collectRuns->total_runs = 0;

	current_record = record_list->next_record;

	while ( current_record != record_list ) {
		if ( current_record->mx_type == MXS_AD_WEDGE ) {
			op_collectRuns->total_runs++;
		}

		current_record = current_record->next_record;
	}

	op_collectRuns->current_run_number = atol( argv[3] );

	mx_free( argv );
	mx_free( operation_args->start_message );

	MX_DEBUG(-2,
	("%s: current_run_number = %lu, total_runs = %lu", fname,
		op_collectRuns->current_run_number,
		op_collectRuns->total_runs));

	if (op_collectRuns->current_run_number >= op_collectRuns->total_runs) {
		/* No runs can be acquired, since the starting run number
		 * is greater than the maximum.
		 */

		op_collectRuns->state = MXBLS_CRS_ERROR;

		return mxbl_broadcast_dcss_operation_error(
			operation_args, MXE_WOULD_EXCEED_LIMIT, fname,
			"The first run requested has a run number (%lu) "
			"that is greater than or equal to the total number "
			"of runs (%lu).", op_collectRuns->current_run_number,
				op_collectRuns->total_runs );
	}

	/* Mark the run sequence as started.  mxblp_update_collectRuns()
	 * is responsible for starting the child runs.
	 */

	op_collectRuns->child_operation_args = NULL;

	op_collectRuns->state = MXBLS_CRS_RUNS_ARE_ACTIVE;

	snprintf( start_message, sizeof(start_message),
		"collecting {Started data collection at run %lu.}",
		op_collectRuns->current_run_number );

	mx_status = mxbl_set_collect_msg( event_manager, start_message );

	MX_DEBUG(-2,("%s complete.", fname));

	return mx_status;
}

static mx_status_type
mxblp_update_collectRuns( MX_CALLBACK_MESSAGE *callback_message,
			MXBL_OPERATION_ARGS *operation_args )
{
	static const char fname[] = "mxblp_update_collectRuns()";

	MXBL_OP_COLLECT_RUNS *op_collectRuns;
	MXBL_OP_COLLECT_RUN *child_collectRun;
	MXBL_EVENT *event;
	MXBL_EVENT_MANAGER *event_manager;
	MXBL_OPERATION_ARGS *child_operation_args;
	char message[100];
	mx_bool_type run_sequence_complete;
	mx_status_type mx_status;

	op_collectRuns = operation_args->op_private;

	event = operation_args->event;

	event_manager = event->event_manager;

	MX_DEBUG(-2,("%s invoked.", fname));

	switch( op_collectRuns->state ) {
	case MXBLS_CRS_RUNS_ARE_ACTIVE:
	    MX_DEBUG(-2,("%s: state = MXBLS_CRS_RUNS_ARE_ACTIVE", fname));

	    child_operation_args = op_collectRuns->child_operation_args;

	    if ( child_operation_args == NULL ) {

		/* If child_operation_args is NULL, then we need to start
		 * a run.  This is done by creating an stog_start_operation
		 * message and then passing it to the event handler for
		 * the collectRun operation.
		 */

		operation_args->client_number = 0;	/* client is 'self' */

		operation_args->operation_number =
			event_manager->operation_number;

		snprintf( event->message_buffer, sizeof(event->message_buffer),
		"stog_start_operation collectRun 0.%lu %lu self",
			event_manager->operation_number,
			op_collectRuns->current_run_number );

		mx_atomic_increment( event_manager->operation_number );

		(void) mxbl_broadcast_dcss_message( event_manager,
							event->message_buffer );

		mx_status = mxbl_start_operation( event,
				&(op_collectRuns->child_operation_args) );

		if ( mx_status.code != MXE_SUCCESS )
			return mx_status;
	    } else {
	    	/* See if the child operation has completed. */

		MX_DEBUG(-2,
		("%s: child_operation_args->operation_in_progress = %d",
			fname, child_operation_args->operation_in_progress));

		if ( child_operation_args->operation_in_progress ) {

			/* Do nothing if the child run is still in progress. */
		} else {
			/* The child run has completed. */

			child_collectRun = child_operation_args->op_private;

			if ( child_collectRun == NULL ) {
				return mxbl_broadcast_dcss_operation_error(
				child_operation_args,
				MXE_CORRUPT_DATA_STRUCTURE, fname,
				"The op_private pointer is NULL for "
				"child operation collectRun %lu.%lu used by "
				"operation collectRuns %lu.%lu",
					child_operation_args->client_number,
					child_operation_args->operation_number,
					operation_args->client_number,
					operation_args->operation_number );
			}

			/* Delete the child's callback. */

			mx_status = mx_function_delete_callback(
					child_collectRun->callback_message );

			if ( mx_status.code != MXE_SUCCESS )
				return mx_status;

			op_collectRuns->child_operation_args = NULL;

			run_sequence_complete = FALSE;

			if ( op_collectRuns->current_run_number == 0 ) {

				/* If the run just completed was run 0,
				 * then we stop here, since run 0 is
				 * never part of a multirun sequence.
				 */

				run_sequence_complete = TRUE;
			} else {
				/* Compute the next run number. */

				op_collectRuns->current_run_number++;

				if ( op_collectRuns->current_run_number >
					op_collectRuns->total_runs )
				{
					/* If we have completed the last run,
					 * then the run sequence is over.
					 */

					run_sequence_complete = TRUE;
				}
			}

			if ( run_sequence_complete ) {

				(void) mxbl_set_collect_msg( event_manager,
			    "inactive {Data collection completed normally.}" );

				op_collectRuns->state =
					MXBLS_CRS_RUNS_ARE_COMPLETE;
				
				operation_args->operation_in_progress = FALSE;

				(void) mxbl_set_collect_msg( event_manager,
			    "inactive {Data collection completed normally.}" );

				snprintf( message, sizeof(message),
				"stog_operation_completed collectRuns "
				"%lu.%lu normal",
					operation_args->client_number,
					operation_args->operation_number );

				(void) mxbl_broadcast_dcss_message(
						event_manager, message );
			}
		}
	    }
	    break;
	}

	return MX_SUCCESSFUL_RESULT;
}

/*==================================================================*/

/* Values for the 'state' variable for the 'detector_collect_image' operation.*/

#define MXBLS_CI_ERROR				(-1)
#define MXBLS_CI_PREPARING_FOR_ACQUISITION	0
#define MXBLS_CI_READY_FOR_MOVE			1
#define MXBLS_CI_MOVING				2
#define MXBLS_CI_MOVE_COMPLETE			3
#define MXBLS_CI_WAITING_FOR_TRANSFER		4
#define MXBLS_CI_TRANSFER_COMPLETE		5
#define MXBLS_CI_WAITING_FOR_READY		6
#define MXBLS_CI_IMAGE_SEQUENCE_COMPLETE	7

typedef struct {
	int dark_cache_number;
	char filename[MXU_FILENAME_LENGTH+1];
	char directory[MXU_FILENAME_LENGTH+1];
	char user_name[MXU_USERNAME_LENGTH+1];
	char motor_name[MXU_BLUICE_NAME_LENGTH+1];
	double time;
	double start_angle;
	double delta;
	double next_position;
	double detector_z;
	double wavelength;
	double detector_horz;
	double detector_vert;
	int mode_index;
	int reuse_dark;

	MX_RECORD *area_detector_record;
	MX_RECORD *motor_record;

	MX_RECORD *detector_z_record;
	MX_RECORD *wavelength_record;
	MX_RECORD *detector_horz_record;
	MX_RECORD *detector_vert_record;

	int state;

	int current_oscillation;
	int number_of_oscillations;

	int number_of_dark_current_frames;
	MX_IMAGE_FRAME **dark_current_frame_array;
} MXBL_OP_DETECTOR_COLLECT_IMAGE;

static mx_status_type mxblp_start_detector_collect_image(MX_CALLBACK_MESSAGE *,
							MXBL_OPERATION_ARGS *);

static mx_status_type mxblp_update_detector_collect_image(MX_CALLBACK_MESSAGE *,
							MXBL_OPERATION_ARGS *);

/*----*/

static mx_status_type
mxbl_op_detector_collect_image( MX_CALLBACK_MESSAGE *callback_message )
{
	static const char fname[] = "mxbl_op_detector_collect_image()";

	MXBL_OPERATION_ARGS *operation_args;
	MX_VIRTUAL_TIMER *oneshot_timer;
	double callback_interval;
	mx_status_type mx_status;

	MX_DEBUG(-2,("%s invoked", fname));

	if ( callback_message == (MX_CALLBACK_MESSAGE *) NULL ) {
		return mx_error( MXE_INVALID_CALLBACK, fname,
		"The MX_CALLBACK_MESSAGE pointer passed was NULL." );
	}

	operation_args = callback_message->u.function.callback_args;

	if ( operation_args == (MXBL_OPERATION_ARGS *) NULL ) {
		return mx_error( MXE_INVALID_CALLBACK, fname,
		"The callback args pointer for callback message %p is NULL.",
			callback_message );
	}

	oneshot_timer = callback_message->u.function.oneshot_timer;

	/* Has the op_detector_collect_image structure been initialized yet? */

	if ( operation_args->op_private == NULL ) {
		mx_status = mxblp_start_detector_collect_image(
				callback_message, operation_args );
	} else {
		mx_status = mxblp_update_detector_collect_image(
				callback_message, operation_args );
	}

	if ( mx_status.code != MXE_SUCCESS ) {
		operation_args->operation_in_progress = FALSE;
	}

	MX_DEBUG(-2,("%s: operation_in_progress = %d",
		fname, (int) operation_args->operation_in_progress));

	/* If the operation is still in progress, reschedule the callback. */

	if ( operation_args->operation_in_progress ) {
		callback_interval =
			callback_message->u.function.callback_interval;

		mx_status = mx_virtual_timer_start( oneshot_timer,
						callback_interval );
	} else {
		mx_status = mx_function_delete_callback( callback_message );
	}

	return mx_status;
}

static mx_status_type
mxblp_start_detector_collect_image( MX_CALLBACK_MESSAGE *callback_message,
					MXBL_OPERATION_ARGS *operation_args )
{
	static const char fname[] = "mxblp_start_detector_collect_image()";

	MXBL_OP_DETECTOR_COLLECT_IMAGE *op_collect_image;
	MXBL_EVENT_MANAGER *event_manager;
	int argc;
	char **argv;
	MX_RECORD *area_detector_record;
	int i, num_dark_frames;
	MX_IMAGE_FRAME **dark_frame_array;
	mx_status_type mx_status;

	/* FIXME: Get rid of this global variable! */

	global_detector_collect_image_args = operation_args;

	event_manager = operation_args->event->event_manager;

	area_detector_record = mxblp_find_area_detector( callback_message );

	if ( area_detector_record == (MX_RECORD *) NULL ) {
		return mx_error( MXE_NOT_FOUND, fname,
		"The MX database does not contain any area detector records." );
	}

	MX_DEBUG(-2,("%s invoked for MX area detector '%s'.",
		fname, area_detector_record->name));

	/* Setup a new op_collect_image structure. */

	op_collect_image = calloc( 1, sizeof(MXBL_OP_DETECTOR_COLLECT_IMAGE) );

	if ( op_collect_image == NULL ) {
		return mx_error( MXE_OUT_OF_MEMORY, fname,
			"Ran out of memory trying to allocate an "
			"MXBL_OP_DETECTOR_COLLECT_IMAGE structure." );
	}

	operation_args->op_private = op_collect_image;

	op_collect_image->area_detector_record = area_detector_record;

	op_collect_image->state = MXBLS_CI_PREPARING_FOR_ACQUISITION;

	/* Split up the start message into tokens. */

	MX_DEBUG(-2,("%s: start message = '%s'",
			fname, operation_args->start_message ));

	mx_string_split( operation_args->start_message, " ", &argc, &argv );

	if ( argc != 17 ) {
		mx_free( argv );

		return mx_error( MXE_NETWORK_IO_ERROR, fname,
			"The start_operation message from the Blu-Ice server "
			"had an incorrect number of arguments.  "
			"The correct number of arguments is 17." );
	}

	op_collect_image->dark_cache_number = atoi( argv[3] );

	if ( !mxblp_string_is_null(argv[4]) ) {
		strlcpy( op_collect_image->filename, argv[4],
			sizeof(op_collect_image->filename) );
	}
	if ( !mxblp_string_is_null(argv[5]) ) {
		strlcpy( op_collect_image->directory, argv[5],
			sizeof(op_collect_image->directory) );
	}
	if ( !mxblp_string_is_null(argv[6]) ) {
		strlcpy( op_collect_image->user_name, argv[6],
			sizeof(op_collect_image->user_name) );
	}
	if ( !mxblp_string_is_null(argv[7]) ) {
		strlcpy( op_collect_image->motor_name, argv[7],
			sizeof(op_collect_image->motor_name) );
	}

	op_collect_image->time          = atof( argv[8] );
	op_collect_image->start_angle   = atof( argv[9] );
	op_collect_image->delta         = atof( argv[10] );
	op_collect_image->detector_z    = atof( argv[11] );
	op_collect_image->wavelength    = atof( argv[12] );
	op_collect_image->detector_horz = atof( argv[13] );
	op_collect_image->detector_vert = atof( argv[14] );
	op_collect_image->mode_index    = atoi( argv[15] );
	op_collect_image->reuse_dark    = atoi( argv[16] );

#if 1
	MX_DEBUG(-2,("%s:  dark_cache_number = %d",
		fname, op_collect_image->dark_cache_number));
	MX_DEBUG(-2,("%s:  filename = '%s'",
		fname, op_collect_image->filename));
	MX_DEBUG(-2,("%s:  directory = '%s'",
		fname, op_collect_image->directory));
	MX_DEBUG(-2,("%s:  user_name = '%s'",
		fname, op_collect_image->user_name));
	MX_DEBUG(-2,("%s:  motor_name = '%s'",
		fname, op_collect_image->motor_name));
	MX_DEBUG(-2,("%s:  time = %lf",
		fname, op_collect_image->time));
	MX_DEBUG(-2,("%s:  start_angle = %lf",
		fname, op_collect_image->start_angle));
	MX_DEBUG(-2,("%s:  delta = %lf",
		fname, op_collect_image->delta));
	MX_DEBUG(-2,("%s:  detector_z = %lf",
		fname, op_collect_image->detector_z));
	MX_DEBUG(-2,("%s:  wavelength = %lf",
		fname, op_collect_image->wavelength));
	MX_DEBUG(-2,("%s:  detector_horz = %lf",
		fname, op_collect_image->detector_horz));
	MX_DEBUG(-2,("%s:  detector_vert = %lf",
		fname, op_collect_image->detector_vert));
	MX_DEBUG(-2,("%s:  mode_index = %d",
		fname, op_collect_image->mode_index));
	MX_DEBUG(-2,("%s:  reuse_dark = %d",
		fname, op_collect_image->reuse_dark));
#endif

	/* If reuse_dark == TRUE, then we will only be taking one image.
	 * Otherwise, we will need to take a number of dark current frames
	 * prior to taking the final image.
	 */

	op_collect_image->current_oscillation = 0;

	if ( op_collect_image->reuse_dark ) {
		op_collect_image->number_of_oscillations = 1;
	} else {
		op_collect_image->number_of_oscillations = 4;
	}

	MX_DEBUG(-2,("%s: number_of_oscillations = %d",
		fname, op_collect_image->number_of_oscillations));

	if ( op_collect_image->number_of_oscillations < 1 ) {
		op_collect_image->number_of_dark_current_frames = 0;
		op_collect_image->dark_current_frame_array = NULL;
	} else {
		/* Allocate memory for the dark current frames. */

		num_dark_frames
			= op_collect_image->number_of_oscillations - 1;

		dark_frame_array = calloc(
			op_collect_image->number_of_dark_current_frames,
				sizeof(MX_IMAGE_FRAME *) );

		if ( dark_frame_array == NULL ) {
			return mxbl_send_dhs_operation_error(
			operation_args, MXE_OUT_OF_MEMORY, fname,
				"Ran out of memory trying to allocate "
				"%d MX_IMAGE_FRAME structures.",
					num_dark_frames );
		}

		for ( i = 0; i < num_dark_frames; i++ ) {
			dark_frame_array[i] = NULL;
		}

		op_collect_image->number_of_dark_current_frames
							= num_dark_frames;

		op_collect_image->dark_current_frame_array = dark_frame_array;
	}

	MX_DEBUG(-2,("%s: number_of_dark_current_frames = %d",
		fname, op_collect_image->number_of_dark_current_frames));

	/* If requested, look for the specified motor record. */

	op_collect_image->motor_record = mxblp_find_motor( callback_message,
						op_collect_image->motor_name );

	if ( strlen(op_collect_image->motor_name) == 0 ) {
		/* Do nothing if no motor name was specified. */
	} else
	if ( op_collect_image->motor_record == NULL ) {
		(void) mx_error( MXE_NOT_FOUND, fname,
		"Blu-Ice motor '%s' was not found in the MX database.",
			op_collect_image->motor_name );
	} else {
		MX_DEBUG(-2,("%s: Blu-Ice motor '%s' == MX motor '%s'",
		fname, op_collect_image->motor_name,
		op_collect_image->motor_record->name ));
	}

	mx_free( operation_args->start_message );
	mx_free( argv );

	/* Find the other motor records. */

	op_collect_image->detector_z_record = mxblp_find_motor(
					callback_message, "detector_z" );

	op_collect_image->wavelength_record = mxblp_find_motor(
					callback_message, "wavelength" );

	op_collect_image->detector_horz_record = mxblp_find_motor(
					callback_message, "detector_horz" );

	op_collect_image->detector_vert_record = mxblp_find_motor(
					callback_message, "detector_vert" );
#if 1
	if ( op_collect_image->detector_z_record != NULL ) {
		MX_DEBUG(-2,("%s: Blu-Ice motor 'detector_z' == MX motor '%s'",
		fname, op_collect_image->detector_z_record->name ));
	}
	if ( op_collect_image->wavelength_record != NULL ) {
		MX_DEBUG(-2,("%s: Blu-Ice motor 'wavelength' == MX motor '%s'",
		fname, op_collect_image->wavelength_record->name ));
	}
	if ( op_collect_image->detector_horz_record != NULL ) {
		MX_DEBUG(-2,
			("%s: Blu-Ice motor 'detector_horz' == MX motor '%s'",
			fname, op_collect_image->detector_horz_record->name ));
	}
	if ( op_collect_image->detector_vert_record != NULL ) {
		MX_DEBUG(-2,
			("%s: Blu-Ice motor 'detector_vert' == MX motor '%s'",
			fname, op_collect_image->detector_vert_record->name ));
	}
#endif

	/* Begin moving the motors to the start position. */

	MX_DEBUG(-2,("%s: Moving motors to the start position.",fname));

	if ( op_collect_image->motor_record != NULL ) {

		MX_DEBUG(-2,("%s: moving '%s' to %f", fname,
			op_collect_image->motor_record->name,
			op_collect_image->start_angle));

		mx_status = mx_motor_move_absolute(
				op_collect_image->motor_record,
				op_collect_image->start_angle, 0 );

		if ( mx_status.code != MXE_SUCCESS ) {
			return mxbl_forward_dhs_operation_error(
					operation_args, mx_status );
		}
	}
	if ( op_collect_image->detector_z_record != NULL ) {

		MX_DEBUG(-2,("%s: moving '%s' to %f", fname,
			op_collect_image->detector_z_record->name,
			op_collect_image->detector_z));

		mx_status = mx_motor_move_absolute(
				op_collect_image->detector_z_record,
				op_collect_image->detector_z, 0 );

		if ( mx_status.code != MXE_SUCCESS ) {
			return mxbl_forward_dhs_operation_error(
					operation_args, mx_status );
		}
	}
	if ( op_collect_image->wavelength_record != NULL ) {

		MX_DEBUG(-2,("%s: moving '%s' to %f", fname,
			op_collect_image->wavelength_record->name,
			op_collect_image->wavelength));

		mx_status = mx_motor_move_absolute(
				op_collect_image->wavelength_record,
				op_collect_image->wavelength, 0 );

		if ( mx_status.code != MXE_SUCCESS ) {
			return mxbl_forward_dhs_operation_error(
					operation_args, mx_status );
		}
	}
	if ( op_collect_image->detector_horz_record != NULL ) {

		MX_DEBUG(-2,("%s: moving '%s' to %f", fname,
			op_collect_image->detector_horz_record->name,
			op_collect_image->detector_horz));

		mx_status = mx_motor_move_absolute(
				op_collect_image->detector_horz_record,
				op_collect_image->detector_horz, 0 );

		if ( mx_status.code != MXE_SUCCESS ) {
			return mxbl_forward_dhs_operation_error(
					operation_args, mx_status );
		}
	}
	if ( op_collect_image->detector_vert_record != NULL ) {

		MX_DEBUG(-2,("%s: moving '%s' to %f", fname,
			op_collect_image->detector_vert_record->name,
			op_collect_image->detector_vert));

		mx_status = mx_motor_move_absolute(
				op_collect_image->detector_vert_record,
				op_collect_image->detector_vert, 0 );

		if ( mx_status.code != MXE_SUCCESS ) {
			return mxbl_forward_dhs_operation_error(
					operation_args, mx_status );
		}
	}

	/* Select one-shot mode for the area detector with the 
	 * requested exposure time.
	 */

	MX_DEBUG(-2,
    ("%s: Setting area detector '%s' to one-shot mode with exposure time = %f",
    		fname, area_detector_record->name, op_collect_image->time));

	mx_status = mx_area_detector_set_one_shot_mode( area_detector_record,
							op_collect_image->time);

	if ( mx_status.code != MXE_SUCCESS ) {
		return mxbl_forward_dhs_operation_error(
					operation_args, mx_status );
	}

	op_collect_image->state = MXBLS_CI_MOVING;

	MX_DEBUG(-2,("%s complete.", fname));

	return MX_SUCCESSFUL_RESULT;
}

static mx_status_type
mxblp_update_detector_collect_image( MX_CALLBACK_MESSAGE *callback_message,
					MXBL_OPERATION_ARGS *operation_args )
{
	static const char fname[] = "mxblp_update_detector_collect_image()";

	MXBL_OP_DETECTOR_COLLECT_IMAGE *op_collect_image;
	MXBL_EVENT_MANAGER *event_manager;
	MX_RECORD *area_detector_record;
	MX_AREA_DETECTOR *area_detector;
	MX_SOCKET *mx_socket;
	char message_buffer[500];
	int n;
	double new_destination;
	mx_bool_type busy, busy_seen;
	mx_status_type mx_status;

	op_collect_image = operation_args->op_private;

	event_manager = operation_args->event->event_manager;

	area_detector_record = op_collect_image->area_detector_record;

	area_detector = area_detector_record->record_class_struct;

	mx_socket = operation_args->event->mx_socket;

	MX_DEBUG(-2,("%s invoked for area detector '%s'.",
		fname, area_detector_record->name ));

	switch( op_collect_image->state ) {
	case MXBLS_CI_READY_FOR_MOVE:
		MX_DEBUG(-2,("%s: state = MXBLS_CI_READY_FOR_MOVE", fname));

		/* Start moving the primary motor to the next position. */
	
		if ( op_collect_image->motor_record != NULL ) {

			n = op_collect_image->current_oscillation;

			new_destination = op_collect_image->start_angle
						+ n * op_collect_image->delta;

			MX_DEBUG(-2,("%s: moving '%s' to %f", fname,
				op_collect_image->motor_record->name,
				new_destination));

			mx_status = mx_motor_move_absolute(
					op_collect_image->motor_record,
					new_destination, 0 );
	
			if ( mx_status.code != MXE_SUCCESS ) {
				return mxbl_forward_dhs_operation_error(
						operation_args, mx_status );
			}
		}

		op_collect_image->state = MXBLS_CI_MOVING;
		break;
	
	case MXBLS_CI_MOVING:
		MX_DEBUG(-2,("%s: state = MXBLS_CI_MOVING", fname));

		busy_seen = FALSE;

		if ( op_collect_image->motor_record != NULL ) {
			mx_status = mx_motor_is_busy(
				op_collect_image->motor_record, &busy );

			if ( mx_status.code != MXE_SUCCESS ) {
				return mxbl_forward_dhs_operation_error(
						operation_args, mx_status );
			}
			if ( busy ) {
				busy_seen = TRUE;
			}
		}

		if ( (busy_seen == FALSE)
		  && (op_collect_image->detector_z_record != NULL) )
		{
			mx_status = mx_motor_is_busy(
				op_collect_image->detector_z_record, &busy );

			if ( mx_status.code != MXE_SUCCESS ) {
				return mxbl_forward_dhs_operation_error(
						operation_args, mx_status );
			}
			if ( busy ) {
				busy_seen = TRUE;
			}
		}

		if ( (busy_seen == FALSE)
		  && (op_collect_image->wavelength_record != NULL) )
		{
			mx_status = mx_motor_is_busy(
				op_collect_image->wavelength_record, &busy );

			if ( mx_status.code != MXE_SUCCESS ) {
				return mxbl_forward_dhs_operation_error(
						operation_args, mx_status );
			}
			if ( busy ) {
				busy_seen = TRUE;
			}
		}

		if ( (busy_seen == FALSE)
		  && (op_collect_image->detector_horz_record != NULL) )
		{
			mx_status = mx_motor_is_busy(
				op_collect_image->detector_horz_record, &busy );

			if ( mx_status.code != MXE_SUCCESS ) {
				return mxbl_forward_dhs_operation_error(
						operation_args, mx_status );
			}
			if ( busy ) {
				busy_seen = TRUE;
			}
		}

		if ( (busy_seen == FALSE)
		  && (op_collect_image->detector_vert_record != NULL) )
		{
			mx_status = mx_motor_is_busy(
				op_collect_image->detector_vert_record, &busy );

			if ( mx_status.code != MXE_SUCCESS ) {
				return mxbl_forward_dhs_operation_error(
						operation_args, mx_status );
			}
			if ( busy ) {
				busy_seen = TRUE;
			}
		}

		MX_DEBUG(-2,("%s: busy_seen = %d", fname, (int) busy_seen));

		/* If one or more motors are still busy moving, then we
		 * return without changing state.
		 */

		if ( busy_seen ) {
			return MX_SUCCESSFUL_RESULT;
		}

		MX_DEBUG(-2,
		("%s: The move to the start position is complete.", fname));

		op_collect_image->state = MXBLS_CI_MOVE_COMPLETE;
		break;

	case MXBLS_CI_MOVE_COMPLETE:
		MX_DEBUG(-2,
		("%s: state = MXBLS_CI_MOVE_COMPLETE", fname));

		/* Send a 'start_oscillation' message to the client. */

		snprintf( message_buffer, sizeof(message_buffer),
			"htos_operation_update detector_collect_image %lu.%lu "
			"start_oscillation NULL %f",
				operation_args->client_number,
				operation_args->operation_number,
				0.5 * op_collect_image->time );

		mx_status = mxbl_send_message( mx_socket, message_buffer,
					NULL, 0, MX_BLUICE_PROTOCOL_2 );

		op_collect_image->state = MXBLS_CI_WAITING_FOR_TRANSFER;
		break;

	case MXBLS_CI_WAITING_FOR_TRANSFER:
		MX_DEBUG(-2,
		("%s: state = MXBLS_CI_WAITING_FOR_TRANSFER", fname));

		/* We wait for the 'detector_transfer_image' operation
		 * to change the state of op_collect_image->state.
		 */

		break;

	case MXBLS_CI_TRANSFER_COMPLETE:
		MX_DEBUG(-2,("%s: state = MXBLS_CI_TRANSFER_COMPLETE", fname));

		/* Send a 'prepare_for_oscillation' message to the client. */

		new_destination = op_collect_image->start_angle
			+ ( op_collect_image->current_oscillation
				* op_collect_image->delta );

		snprintf( message_buffer, sizeof(message_buffer),
			"htos_operation_update detector_collect_image %lu.%lu "
			"prepare_for_oscillation %f",
				operation_args->client_number,
				operation_args->operation_number,
				new_destination );

		mx_status = mxbl_send_message( mx_socket, message_buffer,
					NULL, 0, MX_BLUICE_PROTOCOL_2 );

		op_collect_image->state = MXBLS_CI_WAITING_FOR_READY;
		break;

	case MXBLS_CI_WAITING_FOR_READY:
		MX_DEBUG(-2,
		("%s: state = MXBLS_CI_WAITING_FOR_READY", fname));

		/* We wait for the 'detector_oscillation_ready' operation
		 * to change the state of op_collect_image->state.
		 */

		break;

	case MXBLS_CI_IMAGE_SEQUENCE_COMPLETE:
		MX_DEBUG(-2,
		("%s: state = MXBLS_CI_IMAGE_SEQUENCE_COMPLETE",fname));

		/* The area detector finished frame acquisition successfully. */

		operation_args->operation_in_progress = FALSE;

		snprintf( message_buffer, sizeof(message_buffer ),
			"htos_operation_completed detector_collect_image "
			"%lu.%lu normal "
			"Area detector '%s' image sequence complete.",
				operation_args->client_number,
				operation_args->operation_number,
				area_detector_record->name );

		mx_status = mxbl_send_message( mx_socket, message_buffer,
					NULL, 0, MX_BLUICE_PROTOCOL_2 );
		break;
	default:
		return mxbl_send_dhs_operation_error(
			operation_args, MXE_ILLEGAL_ARGUMENT, fname,
			"Illegal 'detector_collect_image' state %d seen "
			"for area detector '%s'.",
				op_collect_image->state,
				area_detector_record->name );
		break;
	}

	return MX_SUCCESSFUL_RESULT;
}

/*==================================================================*/

static mx_status_type
mxbl_op_detector_oscillation_ready( MX_CALLBACK_MESSAGE *callback_message )
{
	static const char fname[] = "mxbl_op_detector_oscillation_ready()";

	MXBL_OPERATION_ARGS *operation_args;
	MXBL_OP_DETECTOR_COLLECT_IMAGE *op_collect_image;
	MXBL_EVENT *event;
	MX_RECORD *area_detector_record;
	MX_SOCKET *mx_socket;
	char message[100];
	mx_status_type mx_status;

	if ( callback_message == (MX_CALLBACK_MESSAGE *) NULL ) {
		return mx_error( MXE_INVALID_CALLBACK, fname,
		"The MX_CALLBACK_MESSAGE pointer passed was NULL." );
	}

	operation_args = callback_message->u.function.callback_args;

	if ( operation_args == (MXBL_OPERATION_ARGS *) NULL ) {
		return mx_error( MXE_INVALID_CALLBACK, fname,
		"The callback args pointer for callback message %p is NULL.",
			callback_message );
	}

	event = operation_args->event;

	mx_socket = event->mx_socket;

	area_detector_record = mxblp_find_area_detector( callback_message );

	if ( area_detector_record == (MX_RECORD *) NULL ) {
		return mx_error( MXE_NOT_FOUND, fname,
		"The MX database does not contain any area detector records.");
	}

	/* FIXME: Get rid of this global variable! */

	MX_DEBUG(-2,("%s: global_detector_collect_image_args = %p",
		fname, global_detector_collect_image_args));

	if ( global_detector_collect_image_args == NULL ) {
		return mxbl_send_dhs_operation_error(
		operation_args, MXE_NOT_READY, fname,
		"The detector is not currently acquiring an image." );
	}

	op_collect_image = global_detector_collect_image_args->op_private;

	if ( op_collect_image == NULL ) {
		return mxbl_send_dhs_operation_error(
		operation_args, MXE_CORRUPT_DATA_STRUCTURE, fname,
			"Although detector_collect_image is currently in "
			"progress, no MXBL_OP_DETECTOR_COLLECT_IMAGE "
			"structure has been allocated for it." );
	}

	/* Record the fact that we are ready for another oscillation. */

	op_collect_image->state = MXBLS_CI_READY_FOR_MOVE;

	snprintf( message, sizeof(message),
	"htos_operation_completed detector_oscillation_ready %lu.%lu normal",
		operation_args->client_number,
		operation_args->operation_number );

	(void) mxbl_send_message( mx_socket, message,
				NULL, 0, MX_BLUICE_PROTOCOL_2 );

	/* The callback does not need to be rescheduled, so we just delete it.*/

	mx_status = mx_function_delete_callback( callback_message );

	return mx_status;
}

/*==================================================================*/

static mx_status_type
mxbl_op_detector_stop( MX_CALLBACK_MESSAGE *callback_message )
{
	static const char fname[] = "mxbl_op_detector_stop()";

	MXBL_OPERATION_ARGS *operation_args;
	MXBL_EVENT_MANAGER *event_manager;
	MX_RECORD *area_detector_record;
	char message[100];
	mx_status_type mx_status;

	if ( callback_message == (MX_CALLBACK_MESSAGE *) NULL ) {
		return mx_error( MXE_INVALID_CALLBACK, fname,
		"The MX_CALLBACK_MESSAGE pointer passed was NULL." );
	}

	operation_args = callback_message->u.function.callback_args;

	if ( operation_args == (MXBL_OPERATION_ARGS *) NULL ) {
		return mx_error( MXE_INVALID_CALLBACK, fname,
		"The callback args pointer for callback message %p is NULL.",
			callback_message );
	}

	event_manager = operation_args->event->event_manager;

	area_detector_record = mxblp_find_area_detector( callback_message );

	if ( area_detector_record == NULL ) {
		return mx_error( MXE_NOT_FOUND, fname,
		"The MX database does not contain any area detector records.");
	}

	MX_DEBUG(-2,("%s invoked for area detector '%s'.",
		fname, area_detector_record->name));

	/* Send the stop command to the area detector. */

	(void) mx_area_detector_stop( area_detector_record );

	/* Send out a message announcing that the detector has been stopped. */

	snprintf( message, sizeof(message),
		"stog_operation_completed detector_stop %lu.%lu normal",
			operation_args->client_number,
			operation_args->operation_number );

	(void) mxbl_broadcast_dcss_message( event_manager, message );

	/* The callback does not need to be rescheduled, so we just delete it.*/

	mx_status = mx_function_delete_callback( callback_message );

	return mx_status;
}

/*==================================================================*/

static mx_status_type
mxbl_op_detector_transfer_image( MX_CALLBACK_MESSAGE *callback_message )
{
	static const char fname[] = "mxbl_op_detector_transfer_image()";

	MXBL_OPERATION_ARGS *operation_args;
	MXBL_OP_DETECTOR_COLLECT_IMAGE *op_collect_image;
	MXBL_EVENT *event;
	MXBL_EVENT_MANAGER *event_manager;
	MX_RECORD *area_detector_record;
	MX_AREA_DETECTOR *area_detector;
	MX_SOCKET *mx_socket;
	MX_IMAGE_FRAME **frame_array;
	char message[500];
	char status[500];
	char output_filename[MXU_FILENAME_LENGTH+1];
	int n, collect_state;
	unsigned long desired_correction_flags, saved_correction_flags;
	mx_status_type mx_status, mx_status2;

	if ( callback_message == (MX_CALLBACK_MESSAGE *) NULL ) {
		return mx_error( MXE_INVALID_CALLBACK, fname,
		"The MX_CALLBACK_MESSAGE pointer passed was NULL." );
	}

	operation_args = callback_message->u.function.callback_args;

	if ( operation_args == (MXBL_OPERATION_ARGS *) NULL ) {
		return mx_error( MXE_INVALID_CALLBACK, fname,
		"The callback args pointer for callback message %p is NULL.",
			callback_message );
	}

	event = operation_args->event;

	event_manager = event->event_manager;

	mx_socket = event->mx_socket;

	area_detector_record = mxblp_find_area_detector( callback_message );

	if ( area_detector_record == (MX_RECORD *) NULL ) {
		return mx_error( MXE_NOT_FOUND, fname,
		"The MX database does not contain any area detector records.");
	}

	area_detector = area_detector_record->record_class_struct;

	if ( area_detector == (MX_AREA_DETECTOR *) NULL ) {
		return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
		"The MX_AREA_DETECTOR pointer for area detector '%s' is NULL.",
			area_detector_record->name );
	}

	MX_DEBUG(-2,("%s invoked for area detector '%s'.",
		fname, area_detector_record->name));

	/* The Blu-Ice 'detector_transfer_image' command is interpreted as
	 * an MX area detector readout command.
	 */

	mx_status = mx_area_detector_readout_frame( area_detector_record, 0 );

	if ( mx_status.code != MXE_SUCCESS ) {
		return mxbl_forward_dhs_operation_error( operation_args,
								mx_status );
	}

	/* Record the fact that the image transfer has taken place. */

	/* FIXME: Get rid of this global variable! */

	MX_DEBUG(-2,("%s: global_detector_collect_image_args = %p",
		fname, global_detector_collect_image_args ));

	op_collect_image = NULL;
	collect_state = MXBLS_CI_ERROR;

	strlcpy( status, "error Unknown error.", sizeof(status) );

	if ( global_detector_collect_image_args == NULL ) {
	    strlcpy( status, "normal", sizeof(status) );
	} else {
	    op_collect_image = global_detector_collect_image_args->op_private;

	    if ( op_collect_image == NULL ) {
		snprintf( status, sizeof(status),
			"error Although detector_collect_image is currently in "
			"progress, no MXBL_OP_DETECTOR_COLLECT_IMAGE structure "
			"has been allocated for it." );
	    } else {
		MX_DEBUG(-2,("%s: Finished oscillation %d", fname, 
			op_collect_image->current_oscillation ));

		(op_collect_image->current_oscillation)++;

		if ( op_collect_image->current_oscillation >= 
		    op_collect_image->number_of_oscillations )
		{
		    op_collect_image->state = MXBLS_CI_IMAGE_SEQUENCE_COMPLETE;
		} else {
		    op_collect_image->state = MXBLS_CI_TRANSFER_COMPLETE;
		}

		collect_state = op_collect_image->state;
	    }
	}

	/* Under most circumstances, we have to do extra work after
	 * the transfer completes.
	 */

	switch( collect_state ) {
	case MXBLS_CI_TRANSFER_COMPLETE:
		MX_DEBUG(-2,
		("%s: Starting dark current image correction.",fname));

		/* Correct the image using only the mask and bias corrections.*/

		desired_correction_flags =
			MXFT_AD_IMAGE_FRAME | MXFT_AD_BIAS_FRAME;

		mx_status = mx_area_detector_get_correction_flags(
				area_detector_record, &saved_correction_flags );

		if ( mx_status.code != MXE_SUCCESS ) {
			return mxbl_forward_dhs_operation_error(
					operation_args, mx_status );
		}

		mx_status = mx_area_detector_correct_frame(
					area_detector_record );

		mx_status2 = mx_area_detector_set_correction_flags(
				area_detector_record, saved_correction_flags );

		if ( mx_status2.code != MXE_SUCCESS ) {
			return mxbl_forward_dhs_operation_error(
					operation_args, mx_status2 );
		}

		if ( mx_status.code != MXE_SUCCESS ) {
			return mxbl_forward_dhs_operation_error(
					operation_args, mx_status );
		}

		/* Copy the corrected frame to the appropriate place
		 * in the dark_current_frame_array.
		 */

		/* current_oscillation was already incremented earlier
		 * in this function.
		 */

		n = op_collect_image->current_oscillation - 1;

		frame_array = op_collect_image->dark_current_frame_array;

		MX_DEBUG(-2,("%s: frame_array = %p", fname, frame_array));
		MX_DEBUG(-2,("%s: frame_array[%d] = %p",
			fname, n, frame_array[n]));
		MX_DEBUG(-2,("%s: &(frame_array[%d]) = %p",
			fname, n, &(frame_array[n])));

		mx_status = mx_image_copy_frame( area_detector->image_frame,
						&(frame_array[n]) );

		if ( mx_status.code != MXE_SUCCESS ) {
			return mxbl_forward_dhs_operation_error(
					operation_args, mx_status );
		}

		MX_DEBUG(-2,
		("%s: Dark current image correction complete.", fname));

		/* Everything went well, so return with a normal status. */

		strlcpy( status, "normal Transfer complete.", sizeof(status) );
		break;

	case MXBLS_CI_IMAGE_SEQUENCE_COMPLETE:
		MX_DEBUG(-2,("%s: Compute a new dark current array.", fname));

		if ( op_collect_image->dark_current_frame_array != NULL ) {

			mx_status =
			    mxblp_compute_new_dark_current_frame(
			    	area_detector,
				op_collect_image->number_of_dark_current_frames,
				op_collect_image->dark_current_frame_array );

			if ( mx_status.code != MXE_SUCCESS ) {
				return mxbl_forward_dhs_operation_error(
						operation_args, mx_status );
			}
		}

		/* Correct the primary image and then write it to a file. */

		MX_DEBUG(-2,("%s: Starting image frame correction.", fname));

		mx_status = mx_area_detector_correct_frame(
					area_detector_record );

		if ( mx_status.code != MXE_SUCCESS ) {
			return mxbl_forward_dhs_operation_error(
					operation_args, mx_status );
		}

		/* Construct the name of the file to save the image to. */

		snprintf( output_filename, sizeof(output_filename),
			"%s/%s.img", op_collect_image->directory,
			op_collect_image->filename );

		MX_DEBUG(-2,("%s: Writing image file '%s'.",
			fname, output_filename));

		mx_status = mxblp_image_write_file( area_detector,
							output_filename );

		if ( mx_status.code != MXE_SUCCESS ) {
			return mxbl_forward_dhs_operation_error(
					operation_args, mx_status );
		}

		MX_DEBUG(-2,("%s: Successfully wrote image file '%s'.",
			fname, output_filename));

		strlcpy( status, "normal Image sequence complete.",
			sizeof(status) );
		break;
	}

	/* Notify the client that the detector has been read out. */

	snprintf( message, sizeof(message),
	"htos_operation_completed detector_transfer_image %lu.%lu %s",
			operation_args->client_number,
			operation_args->operation_number,
			status );

	(void) mxbl_send_message( mx_socket, message,
				NULL, 0, MX_BLUICE_PROTOCOL_2 );

	/* The callback does not need to be rescheduled, so we just delete it.*/

	mx_status = mx_function_delete_callback( callback_message );

	return mx_status;
}

/*==================================================================*/

typedef struct {
	MX_THREAD *scan_thread;
	MX_RECORD *scan_record;
	char *start_of_setup_string;
} MXBL_OP_SCAN_MOTOR;

static mx_status_type mxblp_start_scanMotor( MX_CALLBACK_MESSAGE *,
						MXBL_OPERATION_ARGS * );

static mx_status_type mxblp_update_scanMotor( MX_CALLBACK_MESSAGE *,
						MXBL_OPERATION_ARGS * );

static mx_status_type mxblp_scanMotor_thread( MX_THREAD *thread, void *args );

/*----*/

static mx_status_type mxbl_scanMotor_df_open( MX_DATAFILE * );
static mx_status_type mxbl_scanMotor_df_close( MX_DATAFILE * );
static mx_status_type mxbl_scanMotor_df_add_measurement_to_datafile(
						MX_DATAFILE * );

MX_DATAFILE_FUNCTION_LIST mxbl_scanMotor_df_list = {
	mxbl_scanMotor_df_open,
	mxbl_scanMotor_df_close,
	NULL,
	NULL,
	NULL,
	mxbl_scanMotor_df_add_measurement_to_datafile,
	NULL
};

/*----*/

static mx_status_type
mxbl_op_scanMotor( MX_CALLBACK_MESSAGE *callback_message )
{
	static const char fname[] = "mxbl_op_scanMotor()";

	MXBL_OPERATION_ARGS *operation_args;
	MX_VIRTUAL_TIMER *oneshot_timer;
	double callback_interval;
	mx_status_type mx_status;

	if ( callback_message == (MX_CALLBACK_MESSAGE *) NULL ) {
		return mx_error( MXE_INVALID_CALLBACK, fname,
		"The MX_CALLBACK_MESSAGE pointer passed was NULL." );
	}

	operation_args = callback_message->u.function.callback_args;

	if ( operation_args == (MXBL_OPERATION_ARGS *) NULL ) {
		return mx_error( MXE_INVALID_CALLBACK, fname,
		"The callback args pointer for callback message %p is NULL.",
			callback_message );
	}

	oneshot_timer = callback_message->u.function.oneshot_timer;

	/* Has the op_scanMotor structure been initialized yet? */

	if ( operation_args->op_private == NULL ) {
		mx_status = mxblp_start_scanMotor( callback_message,
							operation_args );
	} else {
		mx_status = mxblp_update_scanMotor( callback_message,
							operation_args );
	}

	if ( mx_status.code != MXE_SUCCESS ) {
		operation_args->operation_in_progress = FALSE;
	}

	MX_DEBUG(-2,("%s: operation_in_progress = %d",
		fname, (int) operation_args->operation_in_progress));

	/* If the operation is still in progress, reschedule the callback. */

	if ( operation_args->operation_in_progress ) {
		callback_interval =
			callback_message->u.function.callback_interval;

		mx_status = mx_virtual_timer_start( oneshot_timer,
						callback_interval );
	} else {
		mx_status = mx_function_delete_callback( callback_message );
	}

	return mx_status;
}

#define FREE_SCAN_MOTOR \
	do {					\
		mx_free(argv);			\
		mx_free(motor_argv);		\
		mx_free(detector_argv);		\
		mx_free(timing_argv);		\
		mx_free(temp_argv);		\
		mx_free(dup_start_message);	\
	} while(0)

static mx_status_type
mxblp_start_scanMotor( MX_CALLBACK_MESSAGE *callback_message,
			MXBL_OPERATION_ARGS *operation_args )
{
	static const char fname[] = "mxblp_start_scanMotor()";

	MXBL_OP_SCAN_MOTOR *op_scanMotor;
	MXBL_EVENT_MANAGER *event_manager;
	MX_RECORD *record_list, *scan_record;
	MX_RECORD *motor_record, *detector_record;
	MX_SCAN *scan;
	int argc, motor_argc, detector_argc, timing_argc, temp_argc;
	char **argv = NULL;
	char **motor_argv = NULL;
	char **detector_argv = NULL;
	char **timing_argv = NULL;
	char **temp_argv = NULL;
	int i, num_motors, num_detectors, num_steps, num_scans;
	double start, end, step_size;
	double integration_time, settling_time, delay_between_scans;
	char *dup_start_message;
	char scan_description[500];
	char buffer[500];
	char start_buffer[100];
	char step_size_buffer[100];
	char num_steps_buffer[100];
	char bluice_scan_name[MXU_RECORD_NAME_LENGTH+1];
	mx_bool_type fast_mode;
	mx_status_type mx_status;

	MX_DEBUG(-2,("%s invoked.", fname));

	strlcpy( bluice_scan_name, "bluice_scan", sizeof(bluice_scan_name) );

	event_manager = operation_args->event->event_manager;

	record_list = event_manager->record_list;

	/* Is somebody already running a scan?
	 *
	 * FIXME: If any scan _anywhere_ is running, this mechanism will
	 * say that a scan is in progress even if that scan does not have
	 * any devices in common.  In particular, there is no way to
	 * check to to see if the scan uses any Blu-Ice devices.  Also,
	 * this does not prevent some other MX process from starting
	 * a scan after _this_ scan starts.
	 */

	mx_status = mx_get_fast_mode( record_list, &fast_mode );

	if ( mx_status.code != MXE_SUCCESS )
		return mx_status;

	if ( fast_mode ) {
		return mx_error( MXE_NOT_VALID_FOR_CURRENT_STATE, fname,
		"You cannot start a motor scan since someone else is "
		"already running a scan." );
	}

	/* Setup a new op_scanMotor structure. */

	op_scanMotor = calloc( 1, sizeof(MXBL_OP_SCAN_MOTOR) );

	if ( op_scanMotor == NULL ) {
		return mx_error( MXE_OUT_OF_MEMORY, fname,
			"Ran out of memory trying to allocate an "
			"MXBL_OP_COLLECT_FRAME structure." );
	}

	operation_args->op_private = op_scanMotor;

	/* Split up the start message into tokens. */

	MX_DEBUG(-2,("%s: start message = '%s'",
		fname, operation_args->start_message ));

	dup_start_message = strdup( operation_args->start_message );

	if ( dup_start_message == NULL ) {
		return mx_error( MXE_OUT_OF_MEMORY, fname,
			"Ran out of memory trying to allocate a copy "
			"of the scanMotor start message." );
	}

	mxbl_tcl_string_split( dup_start_message, &argc, &argv );

	if ( argc < 10 ) {
		FREE_SCAN_MOTOR;

		return mx_error( MXE_NETWORK_IO_ERROR, fname,
		"The command to start the 'scanMotor' operation had fewer (%d) "
		"than the required number (10) of arguments.", argc );
	}

	MX_DEBUG(-2,("%s: user = '%s'", fname, argv[3]));
	MX_DEBUG(-2,("%s: session_id = '%s'", fname, argv[4]));
	MX_DEBUG(-2,("%s: motor definitions = '%s'", fname, argv[5]));
	MX_DEBUG(-2,("%s: detectors = '%s'", fname, argv[6]));
	MX_DEBUG(-2,("%s: filters = '%s'", fname, argv[7]));
	MX_DEBUG(-2,("%s: timing = '%s'", fname, argv[8]));
	MX_DEBUG(-2,("%s: prefixes = '%s'", fname, argv[9]));

	/* Save position of the username in the original start message. */

	op_scanMotor->start_of_setup_string =
		strstr( operation_args->start_message, argv[3] );

	/* Start setting up the MX scan description. */

	snprintf( scan_description, sizeof(scan_description),
		"%s scan linear_scan motor_scan \"\" \"\" 1 ",
		bluice_scan_name );

	/* Split up the timing information.  The times are all
	 * in milliseconds here.
	 */

	mxbl_tcl_string_split( argv[8], &timing_argc, &timing_argv );

	if ( timing_argc < 4 ) {
		FREE_SCAN_MOTOR;

		return mx_error( MXE_ILLEGAL_ARGUMENT, fname,
		"Only %d timing parameters were specified, "
		"however 4 parameters are needed.", timing_argc );
	}

	integration_time    = 0.001 * atof( timing_argv[0] );
	settling_time       = 0.001 * atof( timing_argv[1] );
	num_scans           = atoi( timing_argv[2] );
	delay_between_scans = 0.001 * atof( timing_argv[3] );

	MX_DEBUG(-2,("%s: integration_time = %g", fname, integration_time));
	MX_DEBUG(-2,("%s: settling_time = %g", fname, settling_time));
	MX_DEBUG(-2,("%s: num_scans = %d", fname, num_scans));
	MX_DEBUG(-2,("%s: delay_between_scans = %g",
					fname, delay_between_scans));

	/* Split up the list of motor definitions into the individual
	 * definitions for each motor.
	 */

	mxbl_tcl_string_split( argv[5], &motor_argc, &motor_argv );

	num_motors = 0;

	for ( i = 0; i < motor_argc; i++ ) {
		if ( strlen( motor_argv[i] ) == 0 ) {
			break;
		}
		num_motors++;
	}

	MX_DEBUG(-2,("%s: num_motors = %d", fname, num_motors));

	snprintf( buffer, sizeof(buffer), "%d %d ", num_motors, num_motors );
	strlcat( scan_description, buffer, sizeof(scan_description) );

	start_buffer[0] = '\0';
	step_size_buffer[0] = '\0';
	num_steps_buffer[0] = '\0';

	/* Get the information for each motor. */

	temp_argv = NULL;

	for ( i = 0; i < num_motors; i++ ) {
		mxbl_tcl_string_split( motor_argv[i], &temp_argc, &temp_argv );

		if ( temp_argc < 6 ) {
			FREE_SCAN_MOTOR;

			return mx_error( MXE_ILLEGAL_ARGUMENT, fname,
			"Only %d motor parameters were specified for "
			"motor '%s', however 5 parameters are needed.",
				temp_argc, temp_argv[i] );
		}

		num_steps = atol( temp_argv[1] );
		start     = atof( temp_argv[2] );
		end       = atof( temp_argv[3] );
		step_size = atof( temp_argv[4] );

		MX_DEBUG(-2,
		("%s: motor[%d] name = '%s', num_steps = %d, start = %f, "
		"end = %f, step_size = %f, units = '%s'", fname,
			i, temp_argv[0], num_steps, start,
			end, step_size, temp_argv[5]));

		motor_record = mx_get_record( record_list, temp_argv[0] );

		if ( motor_record == NULL ) {
			mx_status = mx_error( MXE_NOT_FOUND, fname,
				"Motor '%s' was not found in the MX database.",
				temp_argv[0] );

			FREE_SCAN_MOTOR;	/* This frees temp_argv. */
			return mx_status;
		}

		mx_free(temp_argv);

		snprintf( buffer, sizeof(buffer), "%s ", motor_record->name );
		strlcat( scan_description, buffer, sizeof(scan_description) );

		snprintf( buffer, sizeof(buffer), "%g ", start );
		strlcat( start_buffer, buffer, sizeof(start_buffer) );

		snprintf( buffer, sizeof(buffer), "%g ", step_size );
		strlcat( step_size_buffer, buffer, sizeof(step_size_buffer) );

		snprintf( buffer, sizeof(buffer), "%d ", num_steps );
		strlcat( num_steps_buffer, buffer, sizeof(num_steps_buffer) );
	}

	/* Split up the list of detector definitions into the individual
	 * definitions for each detector.
	 */

	mxbl_tcl_string_split( argv[6], &detector_argc, &detector_argv );

	num_detectors = 0;

	for ( i = 0; i < detector_argc; i++ ) {
		if ( strlen( detector_argv[i] ) == 0 ) {
			break;
		}
		num_detectors++;
	}

	MX_DEBUG(-2,("%s: num_detectors = %d", fname, num_detectors));

	snprintf( buffer, sizeof(buffer), "%d ", num_detectors );
	strlcat( scan_description, buffer, sizeof(scan_description) );

	for ( i = 0; i < num_detectors; i++ ) {
		detector_record = mx_get_record(record_list, detector_argv[0]);

		if ( detector_record == NULL ) {
			mx_status = mx_error( MXE_NOT_FOUND, fname,
			"Detector '%s' was not found in the MX database.",
				detector_argv[0] );

			FREE_SCAN_MOTOR;	/* This frees detector_argv. */
			return mx_status;
		}

		snprintf(buffer, sizeof(buffer), "%s ", detector_record->name);
		strlcat(scan_description, buffer, sizeof(scan_description));
	}

	/* We do not need the temporary argv arrays any more. */

	FREE_SCAN_MOTOR;

	/* Append a bunch of parameters. */

	snprintf( buffer, sizeof(buffer),
	"0x0 %g preset_time \"%g timer1\" custom \"\" none \"\" ",
			settling_time, integration_time );

	MX_DEBUG(-2,("%s: buffer = '%s'", fname, buffer));
	MX_DEBUG(-2,("%s: start_buffer = '%s'", fname, start_buffer));
	MX_DEBUG(-2,("%s: step_size_buffer = '%s'", fname, step_size_buffer));
	MX_DEBUG(-2,("%s: num_steps_buffer = '%s'", fname, num_steps_buffer));

	strlcat( scan_description, buffer, sizeof(scan_description) );

	/* Finish by appending the scan range parameters. */

	strlcat( scan_description, start_buffer, sizeof(scan_description) );
	strlcat( scan_description, step_size_buffer, sizeof(scan_description) );
	strlcat( scan_description, num_steps_buffer, sizeof(scan_description) );

	MX_DEBUG(-2,("%s: scan_description = '%s'", fname, scan_description));

	/*----- Now attempt to create the scan record in the database. -----*/

	/* If there is already a scan record by that name, then delete it. */

	scan_record = mx_get_record( record_list, bluice_scan_name );

	if ( scan_record != (MX_RECORD *) NULL ) {
		if ( scan_record->mx_superclass != MXR_SCAN ) {
			return mx_error( MXE_ILLEGAL_ARGUMENT, fname,
			"Attempted to create scan '%s' when a non-scan "
			"record of the same name already exists.",
				bluice_scan_name );
		}

		mx_status = mx_delete_record( scan_record );

		if ( mx_status.code != MXE_SUCCESS )
			return mx_status;
	}

	/* Now create the new record. */

	scan_record = NULL;

	mx_status = mx_create_record_from_description( record_list,
				scan_description, &scan_record, 0 );

	if ( mx_status.code != MXE_SUCCESS )
		return mx_status;

	mx_status = mx_finish_record_initialization( scan_record );

	if ( mx_status.code != MXE_SUCCESS )
		return mx_status;

	/* Suppress the standard MX scan progress display. */

	scan = scan_record->record_superclass_struct;

	scan->scan_flags |= MXF_SCAN_SUPPRESS_PROGRESS_DISPLAY;

	/*--- Redirect the scan output so that it can be sent to Blu-Ice. ---*/

	/* FIXME: At the moment, this is a memory leak. */

	mx_status = mx_scan_set_custom_datafile_handler( scan_record,
						operation_args,
						&mxbl_scanMotor_df_list );

	if ( mx_status.code != MXE_SUCCESS )
		return mx_status;

	/*------------------------------------------------------------------*/

	/* At this point the scan record is ready to be executed.
	 * We run the scan in a separate thread so that the DCSS
	 * process will not be blocked.
	 */

	MX_DEBUG(-2,("%s: *** Start scan thread here ****", fname));

	op_scanMotor->scan_record = scan_record;

	mx_status = mx_thread_create( &(op_scanMotor->scan_thread),
			mxblp_scanMotor_thread, operation_args );

	return mx_status;
}

#undef FREE_SCAN_MOTOR

/*---*/

static mx_status_type
mxblp_update_scanMotor( MX_CALLBACK_MESSAGE *callback_message,
			MXBL_OPERATION_ARGS *operation_args )
{
	static const char fname[] = "mxblp_update_scanMotor()";

	MX_DEBUG(-2,("%s invoked.", fname));

	return MX_SUCCESSFUL_RESULT;
}

/*---*/

static mx_status_type
mxblp_scanMotor_thread( MX_THREAD *thread, void *args )
{
	static const char fname[] = "mxblp_scanMotor_thread()";

	MXBL_OPERATION_ARGS *operation_args;
	MXBL_OP_SCAN_MOTOR *op_scanMotor;
	MX_RECORD *scan_record;
	mx_status_type mx_status;

	operation_args = args;

	if ( operation_args == (MXBL_OPERATION_ARGS *) NULL ) {
		return mx_error( MXE_NULL_ARGUMENT, fname,
		"The thread arguments pointer passed was NULL." );
	}

	op_scanMotor = operation_args->op_private;

	if ( op_scanMotor == (MXBL_OP_SCAN_MOTOR *) NULL ) {
		return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
		"The op_private pointer for this operation is NULL." );
	}

	scan_record = op_scanMotor->scan_record;

	if ( scan_record == (MX_RECORD *) NULL ) {
		return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
		"The scan_record pointer for this operation is NULL." );
	}

	MX_DEBUG(-2,("%s invoked for scan record '%s'.",
		fname, scan_record->name ));

	/* Run the scan. */

	mx_status = mx_perform_scan( scan_record );

	MX_DEBUG(-2,("%s: scan '%s' completed with status = %ld",
		fname, scan_record->name, mx_status.code ));

	operation_args->operation_in_progress = FALSE;

	return MX_SUCCESSFUL_RESULT;
}

/*----*/

static mx_status_type
mxbl_scanMotor_df_open( MX_DATAFILE *datafile )
{
	static const char fname[] = "mxbl_scanMotor_df_open()";

	MX_DATAFILE_CUSTOM *custom_datafile;
	MXBL_OPERATION_ARGS *operation_args;
	MXBL_OP_SCAN_MOTOR *op_scanMotor;
	char message[500];
	mx_status_type mx_status;

	MX_DEBUG(-2,("%s invoked.", fname));

	custom_datafile = datafile->datafile_type_struct;

	operation_args = custom_datafile->custom_args;

	op_scanMotor = operation_args->op_private;

	snprintf( message, sizeof(message),
	"stog_operation_update scanMotor %lu.%lu setup %s",
		operation_args->client_number,
		operation_args->operation_number,
		op_scanMotor->start_of_setup_string );

	mx_status = mxbl_broadcast_dcss_message(
			operation_args->event->event_manager, message );

	return mx_status;
}

static mx_status_type
mxbl_scanMotor_df_close( MX_DATAFILE *datafile )
{
	static const char fname[] = "mxbl_scanMotor_df_close()";

	MX_DATAFILE_CUSTOM *custom_datafile;
	MXBL_OPERATION_ARGS *operation_args;
	MXBL_OP_SCAN_MOTOR *op_scanMotor;
	char message[80];
	mx_status_type mx_status;

	MX_DEBUG(-2,("%s invoked.", fname));

	custom_datafile = datafile->datafile_type_struct;

	operation_args = custom_datafile->custom_args;

	op_scanMotor = operation_args->op_private;

#if 0
	snprintf( message, sizeof(message),
		"stog_operation_update scanMotor %lu.%lu done",
			operation_args->client_number,
			operation_args->operation_number );
#else
	snprintf( message, sizeof(message),
		"stog_operation_completed scanMotor %lu.%lu normal",
			operation_args->client_number,
			operation_args->operation_number );
#endif

	mx_status = mxbl_broadcast_dcss_message(
			operation_args->event->event_manager, message );

	return mx_status;
}

static mx_status_type
mxbl_scanMotor_df_add_measurement_to_datafile( MX_DATAFILE *datafile )
{
	static const char fname[] =
		"mxbl_scanMotor_df_add_measurement_to_datafile()";

	MX_DATAFILE_CUSTOM *custom_datafile;
	MXBL_OPERATION_ARGS *operation_args;
	MXBL_OP_SCAN_MOTOR *op_scanMotor;
	MX_SCAN *scan;
	MX_RECORD *motor_record, *input_record;
	char message[80];
	char buffer[40];
	double position[2];
	double double_value;
	long long_value;
	unsigned long i;
	mx_status_type mx_status;

	MX_DEBUG(-2,("%s invoked.", fname));

	custom_datafile = datafile->datafile_type_struct;

	operation_args = custom_datafile->custom_args;

	op_scanMotor = operation_args->op_private;

	scan = datafile->scan;

	snprintf( message, sizeof(message),
		"stog_operation_update scanMotor %lu.%lu ",
			operation_args->client_number,
			operation_args->operation_number );

	/* Add the motor positions. */

	for ( i = 0; i < scan->num_motors; i++ ) {
		motor_record = scan->motor_record_array[i];

		mx_status = mx_motor_get_position( motor_record,
						&(position[i]) );

		if ( mx_status.code != MXE_SUCCESS )
			return mx_status;
	}

	switch( scan->num_motors ) {
	case 1:
		snprintf( buffer, sizeof(buffer), "%f ", position[0] );
		break;
	case 2:
		snprintf( buffer, sizeof(buffer), "{%f %f} ",
					position[0], position[1] );
		break;
	}

	strlcat( message, buffer, sizeof(message) );

	/* Add the input device values. */

	if ( scan->num_input_devices > 1 ) {
		strlcat( message, "{", sizeof(message) );
	}

	for ( i = 0; i < scan->num_input_devices; i++ ) {
		input_record = scan->input_device_array[i];

		switch( input_record->mx_class ) {
		case MXC_ANALOG_INPUT:
			mx_status = mx_analog_input_read( input_record,
							&double_value );

			if ( mx_status.code != MXE_SUCCESS )
				return mx_status;
			break;

		case MXC_SCALER:
			mx_status = mx_scaler_read( input_record, &long_value );

			if ( mx_status.code != MXE_SUCCESS )
				return mx_status;

			double_value = long_value;
			break;
		}

		if ( i == 0 ) {
			snprintf( buffer, sizeof(buffer), "%f", double_value );
		} else {
			snprintf( buffer, sizeof(buffer), " %f", double_value );
		}

		strlcat( message, buffer, sizeof(message) );
	}

	if ( scan->num_input_devices > 1 ) {
		strlcat( message, "}", sizeof(message) );
	}

	mx_status = mxbl_broadcast_dcss_message(
			operation_args->event->event_manager, message );

	return mx_status;
}

/*==================================================================*/

mx_status_type
mxbl_op_default_destructor( MX_CALLBACK_MESSAGE *callback_message )
{
	MXBL_OPERATION_ARGS *operation_args;

	if ( callback_message == NULL ) {
		return MX_SUCCESSFUL_RESULT;
	}

	operation_args = callback_message->u.function.callback_args;

	if ( operation_args == NULL ) {
		return MX_SUCCESSFUL_RESULT;
	}

	mx_free( operation_args->op_private );
	mx_free( operation_args );

	/* Note: The MX_CALLBACK_MESSAGE structure will be freed
	 * by mx_function_delete_callback() directly.
	 */

	return MX_SUCCESSFUL_RESULT;
}

/*------------------------------------------------------------------*/

mx_status_type
mxbl_op_detector_collect_image_destructor(
		MX_CALLBACK_MESSAGE *callback_message )
{
	MXBL_OP_DETECTOR_COLLECT_IMAGE *op_collect_image;
	int i, num_dark_frames;
	MX_IMAGE_FRAME **dark_frame_array;
	MX_IMAGE_FRAME *dark_frame;
	mx_status_type mx_status;

	/* FIXME: Get rid of this global variable! */

	if ( global_detector_collect_image_args != NULL ) {
	    op_collect_image = global_detector_collect_image_args->op_private;

	    if ( op_collect_image != NULL ) {
	    	num_dark_frames =
			op_collect_image->number_of_dark_current_frames;

		dark_frame_array =
			op_collect_image->dark_current_frame_array;

		if ( dark_frame_array != NULL ) {

		    for ( i = 0; i < num_dark_frames; i++ ) {
		    	dark_frame = dark_frame_array[i];

			if ( dark_frame != NULL ) {
			    mx_image_free( dark_frame );
			}
		    }

		    mx_free( dark_frame_array );
		}
	    }
	}

	global_detector_collect_image_args = NULL;

	mx_status = mxbl_op_default_destructor( callback_message );

	return mx_status;
}

