/*
 * Name:    mxbl_util.c
 *
 * Purpose: Utility functions for the MX Blu-Ice server.
 *
 * Author:  William Lavender
 *
 *--------------------------------------------------------------------------
 *
 * Copyright 2008 Illinois Institute of Technology
 *
 * See the file "LICENSE" for information on usage and redistribution
 * of this file, and for a DISCLAIMER OF ALL WARRANTIES.
 *
 */

#define MXBL_UTIL_DEBUG		TRUE

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "mx_util.h"
#include "mx_record.h"
#include "mx_list.h"
#include "mx_callback.h"

#include "mx_poison.h"

#include "mxbl.h"

/*------------------------------------------------------------------*/

#define MXBLT_NOT_IN_TOKEN	0
#define MXBLT_BRACE_TOKEN	1
#define MXBLT_STRING_TOKEN	2

mx_status_type
mxbl_tcl_string_split( char *tcl_string,
			int *argc,
			char ***argv )
{
	static const char fname[] = "mxbl_tcl_string_split()";

	unsigned long block_size, num_blocks, array_size;
	char c;
	int brace_level, token_type;
	unsigned long i, length;
	mx_bool_type exit_loop;

	if ( (tcl_string == NULL)
	  || (argc == NULL)
	  || (argv == NULL) )
	{
		return mx_error( MXE_NULL_ARGUMENT, fname,
		"One or more of the function arguments was NULL." );
	}

#if MXBL_UTIL_DEBUG
	MX_DEBUG(-2,("%s invoked for string '%s'", fname, tcl_string));
#endif

	block_size = 10;
	num_blocks = 1;

	array_size = num_blocks * block_size;

	*argv = malloc( array_size * sizeof(char **) );

	if ( *argv == NULL ) {
		return mx_error( MXE_OUT_OF_MEMORY, fname,
		"Unable to allocate memory for a %lu element argv array.",
			array_size );
	}

	*argc = 0;
	brace_level = 0;
	token_type = MXBLT_NOT_IN_TOKEN;
	exit_loop = FALSE;

	length = strlen(tcl_string);

	for ( i = 0; i < length; i++ ) {
		/* Do we need to expand the size of argv? */

		if ( (*argc) >= array_size ) {
			num_blocks++;
			array_size = num_blocks * block_size;

			*argv = realloc( *argv, array_size * sizeof(char**) );

			if ( *argv == NULL ) {
				return mx_error( MXE_OUT_OF_MEMORY, fname,
				"Unable to increase the size of the argv "
				"array to %lu elements.", array_size );
			}
		}

		/* Look at the next character in the string. */

		c = tcl_string[i];

#if 0 && MXBL_UTIL_DEBUG
		MX_DEBUG(-2,("tcl_string[%lu] = %#x '%c'", i, c, c));
#endif

		if ( token_type == MXBLT_NOT_IN_TOKEN ) {
			switch(c) {
			case ' ':
			case '\t':
				/* We are still not in a token, so do nothing.*/
				break;

			case '\n':
				/* We have reached the end of the line. */

				tcl_string[i] = '\0';
				exit_loop = TRUE;
				break;

			case '{':
				/* This is the start of a brace token. */

				tcl_string[i] = '\0';

				(*argv)[*argc] = &(tcl_string[i+1]);

				(*argc)++;

				brace_level++;

				token_type = MXBLT_BRACE_TOKEN;
				break;

			default:
				/* This is the start of a string token. */

				(*argv)[*argc] = &(tcl_string[i]);

				(*argc)++;

				token_type = MXBLT_STRING_TOKEN;
				break;
			}
		} else
		if ( token_type == MXBLT_BRACE_TOKEN ) {
			switch(c) {
			case '{':
				brace_level++;
				break;
			case '}':
				brace_level--;
				break;
			}

			if ( brace_level < 0 ) {
				return mx_error( MXE_UNPARSEABLE_STRING, fname,
					"Invalid brace level %d at "
					"string offset %lu.",
						brace_level, i );
			} else
			if ( brace_level == 0 ) {
				tcl_string[i] = '\0';
				token_type = MXBLT_NOT_IN_TOKEN;
			}
		} else
		if ( token_type == MXBLT_STRING_TOKEN ) {
			switch(c) {
			case ' ':
			case '\t':
			case '\n':
				tcl_string[i] = '\0';
				token_type = MXBLT_NOT_IN_TOKEN;
			}
		} else {
			return mx_error( MXE_ILLEGAL_ARGUMENT, fname,
			"Illegal token type %d seen at string offset %lu.",
				token_type, i );
		}

#if 0 && MXBL_UTIL_DEBUG
		MX_DEBUG(-2,
		("argc = %d, token_type = %d, brace_level = %d, exit_loop = %d",
			*argc, token_type, brace_level, exit_loop));
#endif

		if ( exit_loop ) {
			break;
		}
	}

#if MXBL_UTIL_DEBUG

	MX_DEBUG(-2,("%s: argc = %d", fname, *argc));

	for ( i = 0; i < (*argc); i++ ) {
		MX_DEBUG(-2,("%s: argv[%lu] = '%s'", fname, i, (*argv)[i] ));
	}
#endif

	return MX_SUCCESSFUL_RESULT;
}

